{$X+,R-,S-,D+}

unit Zar;

interface

uses
	Dos, ByDos,
	Objects, Drivers, Memory, Views, Menus, Dialogs, StdDlg, MsgBox, App,
	Gadgets, ZarInc, Zarhlp, Other, HelpFile, Calc, Cale, FileServ, LCR,
	WorkInfo, WorkRez, WorkTab, WorkMain,
	WorkBlis, WorkFond, WorkChisl, WorkSub, WorkAdd;

type

	PZar = ^TZar;
	TZar = Object (TApplication)
		Clock		: PClockView;
		Heap		: PHeapView;
		constructor	Init;
		procedure	GetEvent (Var Event : TEvent); virtual;
		procedure	HandleEvent (Var Event : TEvent); virtual;
		procedure	Idle; virtual;
		procedure	InitMenuBar; virtual;
		procedure	InitStatusLine; virtual;
		procedure	OutOfMemory; virtual;
		procedure	InitDeskTop; virtual;
		function	Valid (Command : Word) : Boolean; virtual;
	end;

implementation

constructor TZar.Init;
	var
		R		: TRect;
		I		: Integer;
		FileName: PathStr;
	begin
		RegisterObjects;
		RegisterViews;
		RegisterMenus;
		RegisterDialogs;
		RegisterApp;
		RegisterCalc;
		RegisterHelpFile;
		RegisterCalendar;
		RegisterAll;

		TApplication.Init;
		GetExtent (R); R.B.Y := 1;
		Insert (New (PView, Init (R)));

		GetExtent (R);
		Inc (R.A.Y);
		R.A.X := R.B.X - 6; R.B.Y := R.A.Y + 1;
		Clock := New(PClockView, Init(R));
		Insert(Clock);

		GetExtent(R);
		Dec(R.B.X);
		R.A.X := R.B.X - 6; R.A.Y := R.B.Y - 1;
		Heap := New (PHeapView, Init(R));
		Insert(Heap);

		DisableCommands ([cmSave,
			cmWorkInfo, cmWorkTabel, cmWorkRez,
			cmSVadd, cmSVsub, cmFond, cmBList, cmChisl]);

		if FileExist (DskName) then
			RetrieveDesktop
		else
			About;
		for I := 1 to ParamCount do
			begin
				FileName := ParamStr(I);
				if FileName [Length(FileName)] = '\' then
					FileName := FileName + '*.ZAR';
				if Pos ('.', FileName) = 0 then
					FileName := FileName + '.ZAR';
				if Pos ('?', FileName) = Pos ('*', FileName) then
					OpenBase (FileName)
				else
					FileOpen (FileName);
			end;
	end;

procedure TZar.InitDeskTop;
	var
		R : TRect;
	begin
		GetExtent (R);
		Inc (R.A.Y, 2);
		Dec (R.B.Y);
		New (DeskTop, Init (R));
	end;

function TZar.Valid (Command : Word) : Boolean;
	begin
		Valid := TApplication.Valid (Command){ and Main.Valid (Command)};
	end;

procedure TZar.GetEvent (Var Event: TEvent);
	var
		W			: PWindow;
		HFile		: PHelpFile;
		HelpStrm	: PDosStream;
	const
		HelpInUse	: Boolean = False;
		ShellInUse	: Boolean = False;
	begin
		TApplication.GetEvent (Event);
		case Event.What of
			evCommand:
				if (Event.Command = cmHelp) and not HelpInUse then
					begin
						HelpInUse := True;
						HelpStrm := New (PDosStream, Init (CalcHelpName, stOpenRead));
						HFile := New (PHelpFile, Init (HelpStrm));
						if HelpStrm^.Status <> stOk then
							begin
								MessageBox (#3'    䠩 .', nil, mfError + mfCancelButton);
								Dispose (HFile, Done);
							end
						else
							begin
								W := New (PHelpWindow, Init (HFile, GetHelpCtx));
								if ValidView (W) <> nil then
									begin
										ExecView (W);
										Dispose(W, Done);
									end;
							end;
						ClearEvent (Event);
						HelpInUse := False;
					end
				else
					if (Event.Command = cmDosShell) and not ShellInUse then
						begin
							ShellInUse := True;
							DosShell ('Norton Commander', 'NCMAIN.EXE', True, True );
							ShellInUse := False;
							ClearEvent (Event);
						end;
			evMouseDown:
				if Event.Buttons <> 1 then Event.What := evNothing;
			end;
	end;

procedure TZar.HandleEvent (Var Event: TEvent);
	function Exist (P : PView) : Boolean;
		begin
			if (PtrRec (PMain).Seg > 0) and (PtrRec (P).Seg > 0) then
				begin
					P^.Select;
					Exist := True;
				end
			else
				Exist := False;
		end;
	var
		Control		: PView;
		HFile		: PHelpFile;
		HelpStrm	: PDosStream;
	begin
		TApplication.HandleEvent(Event);
		case Event.What of
			evCommand:
				case Event.Command of
					cmOpenMain:
						if (PMain = nil) or PMain^.Valid (cmClose) then
							begin
								PMain^.Free;
								FileOpen ('*.ZAR');
							end;
					cmWorkInfo:
						if not Exist (PMain^.Info) then	InsertWindow (New (PWorkInfo, Init (PMain)));
					cmWorkRez:
						if not Exist (PMain^.Rez) then InsertWindow (New (PWorkRez, Init (PMain)));
					cmWorkTabel:
						if not Exist (PMain^.Tabel) then InsertWindow (New (PWorkTabel, Init (PMain)));
					cmBList:
						if not Exist (PMain^.BList) then InsertWindow (New (PWorkBList, Init (PMain)));
					cmFond:
						if not Exist (PMain^.Fond) then InsertWindow (New (PWorkFond, Init (PMain)));
					cmChisl:
						if not Exist (PMain^.Chisl) then InsertWindow (New (PWorkChisl, Init (PMain)));
					cmSVsub:
						if not Exist (PMain^.SVsub) then InsertWindow (New (PWorkSVsub, Init (PMain)));
					cmSVadd:
						if not Exist (PMain^.SVadd) then InsertWindow (New (PWorkSVadd, Init (PMain)));
					cmCloseMain: if (PMain <> nil) and PMain^.Valid (cmClose) then PMain^.Free;
					cmChDir: ChangeDir;
					cmCascade: Cascade;
					cmTile: Tile;
					cmAbout: About;
					cmCalendar: Calendar;
					cmCalculator: Calculator;
					cmSaveDesktop: SaveDesktop;
					cmRetrieveDesktop: RetrieveDesktop;
					cmHelp:	if not HelpInUse then
					begin
						HelpStrm := New (PDosStream, Init (CalcHelpName, stOpenRead));
						HFile := New (PHelpFile, Init (HelpStrm));
						if HelpStrm^.Status <> stOk then
							begin
								MessageBox (#3'    䠩 .', nil, mfError + mfCancelButton);
								HFile^.Free;
							end
						else
							if DeskTop^.Current^.State and sfModal = 0 then
								InsertWindow (New (PHelpWindow, Init (HFile, GetHelpCtx)))
							else
								begin
									Control := New (PHelpWindow, Init (HFile, GetHelpCtx));
									if ValidView (Control) <> nil then
										begin
											ExecView (Control);
											Control^.Free;
										end;
								end;
					end
				else
					Exit;
				end;
			else
				Exit;
			end;
		ClearEvent (Event);
	end;

procedure TZar.Idle;
	function IsTileable(P: PView): Boolean; far;
		begin
			IsTileable := P^.Options and ofTileable <> 0;
		end;
	begin
		TApplication.Idle;
		Clock^.Update;
		Heap^.Update;
		if Desktop^.FirstThat (@IsTileable) <> nil then
			EnableCommands ([cmTile, cmCascade])
		else
			DisableCommands ([cmTile, cmCascade]);
	end;

procedure TZar.InitMenuBar;
	begin
		MenuBar := MakeMenuBar;
	end;

procedure TZar.InitStatusLine;
	begin
		StatusLine := MakeStatusLine;
	end;

procedure TZar.OutOfMemory;
	begin
		MessageBox (#3'筮 ᢮ '#13+
					#3' 믮 .',
					nil, mfError + mfCancelButton);
	end;
end.
