{1/2 Copyright by Leonid Youriev, 1993, os@imfiko.bishkek.su}
{$I VERSION}

unit DataColl;

interface

uses	Objects;

{$I VERSION}

type

	KeyTypes = (StringKey, LongIntKey);

	PDataCollection = ^TDataCollection;
	TDataCollection = Object (TStringCollection)
		ItemSize	: Word;
		KeyType		: KeyTypes;
		Status		: Integer;
		constructor	Init (ALimit, ADelta, AnItemSize : Integer; AKeyType : KeyTypes);
		constructor	Load (Var S : TStream);
		function	Compare (Key1, Key2 : Pointer) : Integer; virtual;
		procedure	Error (Code, Info : Integer); virtual;
		procedure	FreeItem (Item : Pointer); virtual;
		function	GetItem (Var S : TStream) : Pointer; virtual;
		procedure	PutItem (Var S : TStream; Item : Pointer); virtual;
		procedure	SetLimit (ALimit : Integer); virtual;
		procedure	Store (Var S : TStream); virtual;
	end;

const
	RDataCollection : TStreamRec = (
		ObjType	: 10050;
		VmtLink	: Ofs (TypeOf (TDataCollection)^);
		Load	: @TDataCollection.Load;
		Store	: @TDataCollection.Store);

procedure RegisterDataColl;

implementation

uses	Memory, CharLog, Quick;

procedure RegisterDataColl;
begin
	RegisterType (RDataCollection);
end;

constructor TDataCollection.Init (ALimit, ADelta, AnItemSize : Integer; AKeyType : KeyTypes);
begin
	TStringCollection.Init (ALimit, ADelta);
	ItemSize := AnItemSize;
	KeyTYpe := AKeyType;
end;

constructor TDataCollection.Load (Var S : TStream);
begin
	S.Read (ItemSize, SizeOf (ItemSize));
	TStringCollection.Load (S);
	S.Read (KeyType, SizeOf (KeyType));
	Status := 0;
end;

function TDataCollection.Compare (Key1, Key2 : Pointer) : Integer;
Var
	SK1,
	SK2		: String;
	I		: Integer;
begin
	if KeyType = StringKey then
		begin
			SK1 := PString (Key1)^; UpStr (SK1);
			SK2 := PString (Key2)^; UpStr (SK2);
			Compare := TStringCollection.Compare (@SK1, @SK2);
		end
	else
		if KeyType = LongIntKey then
			Compare := SignLong (Longint (Key1^) - Longint (Key2^));
end;

procedure TDataCollection.Error (Code, Info : Integer);
begin
	Status := Code;
end;

procedure TDataCollection.FreeItem (Item : Pointer);
begin
	if Item <> nil then FreeMem (Item, ItemSize);
end;

function TDataCollection.GetItem (Var S : TStream) : Pointer;
Var
	Item	: Pointer;
begin
	GetMem (Item, ItemSize);
	S.Read (Item^, ItemSize);
	GetItem := Item;
end;

procedure TDataCollection.PutItem (Var S : TStream; Item : Pointer);
begin
	S.Write (Item^, ItemSize);
end;

procedure TDataCollection.SetLimit (ALimit : Integer);
Var
	AItems	: PItemList;
begin
	if ALimit < Count then ALimit := Count;
	if ALimit > MaxCollectionSize then ALimit := MaxCollectionSize;
	if ALimit <> Limit then
		begin
			if ALimit = 0 then
				AItems := nil
			else
				begin
					AItems := MemAlloc (Word (ALimit) * SizeOf (Pointer));
					if AItems = nil then Exit;
					if Count <> 0 then Move (Items^, AItems^, Word (Count) * SizeOf (Pointer));
				end;
			if Limit <> 0 then FreeMem (Items, Word (Limit) * SizeOf (Pointer));
			Items := AItems;
			Limit := ALimit;
		end;
end;

procedure TDataCollection.Store (Var S : TStream);
begin
	S.Write (ItemSize, SizeOf (ItemSize));
	TStringCollection.Store (S);
	S.Write (KeyType, SizeOf (KeyType));
end;

end.
