{Copyright by Leonid Youriev, 1993, os@imfiko.bishkek.su}
{$I VERSION}

unit DView;

interface

uses	Objects, TabView;

{$I VERSION}

type

	PCFloat = ^TCFloat;
	TCFloat = Object (TColumn)
		DecWid	: Integer;
		constructor Init (ATitle : PString; AWidth, ADecWid, AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
	end;

	PCString = ^TCString;
	TCString = Object (TColumn)
		constructor Init (ATitle : PString; AWidth, AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
	end;

	PCInteger = ^TCInteger;
	TCInteger = Object (TColumn)
		constructor Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
	end;

	PCYesNo = ^TCYesNo;
	TCYesNo = Object (TColumn)
		constructor Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
		procedure	NextState (By : Boolean; Var Data); virtual;
	end;

	PCByte = ^TCByte;
	TCByte = Object (TColumn)
		constructor Init (ATitle : PString; AWidth, AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
	end;

	PCLongint = ^TCLongint;
	TCLongint = Object (TColumn)
		constructor Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
	end;

	PMobileRec = ^TMobileRec;
	TMobileRec = Object (TObject)
		Next	: PMobileRec;
		Name	: PString;
		constructor Init (AName : PString; ANext : PMobileRec);
		destructor	Done; virtual;
	end;

	PCMobile = ^TCMobile;
	TCMobile = Object (TColumn)
		Base	: PMobileRec;
		Max		: Integer;
		constructor	Init (ATitle : PString; ABase : PMobileRec; AHeight : Integer; ANext : PColumn);
		procedure	GetView (Var S : String; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
		procedure	NextState (By : Boolean; Var Data); virtual;
		destructor	Done; virtual;
	end;

	procedure ForNumber (Var S : String);
	procedure ForMobile (Var S : String);

implementation

uses	CharLog;

{***************************************************************** TCYesNo **}
constructor TCMobile.Init (ATitle : PString; ABase : PMobileRec; AHeight : Integer; ANext : PColumn);
	Var
		AMax	: Integer;
		AWidth	: Integer;
		P		: PMobileRec;
	begin
		AMax := 0;
		AWidth := 0;
		P := ABase;
		while PtrRec (P).Seg > 0 do
			begin
				if AWidth < Length (P^.Name^) then AWidth := Length (P^.Name^);
				P := P^.Next;
				Inc (AMax);
			end;
		TColumn.Init (ATitle, SizeOf (Byte), AWidth, AHeight, nil, ANext);
		Flags := Flags or cfMobile;
		Base := ABase;
		Max := AMax;
	end;

procedure TCMobile.GetView (Var S : String; Var Data);
	Var
		I : Byte;
		P : PMobileRec;
	begin
		I := Byte (Data);
		if I >= Max then
			S := '???'
		else
			begin
				P := Base;
				while (I > 0) do
					begin
						P := P^.Next;
						Dec (I);
					end;
				S := P^.Name^;
			end;
		I := (Width - Length (S)) shr 1;
		S := Space (I) + S + Space (Width - I - Length (S));
	end;

function Cmp (Max : Integer; Var S1, S2 : String) : Boolean;
	Var
		I : Integer;
	begin
		Cmp := False;
		if Max > Length (S1) then Max := Length (S1);
		if Max > Length (S2) then Max := Length (S2);
		for I := 1 to Max do
			if S2 [I] <> AlfaUpCase (S1 [I]) then Exit;
		Cmp := True;
	end;

function TCMobile.SetView (S : String; Var Data) : Integer;
	Var
		W, I, N : Integer;
		P		: PMobileRec;
		B		: Byte;
	begin
		SetView := 1;
		ForMobile (S);
		for W := 1 to Width - 1 do
			begin
				P := Base;
				N := 0; I := 0;
				while (PtrRec (P).Seg > 0) and (N < 2) do
					begin
						if Cmp (W, P^.Name^, S) then
							begin
								Inc (N);
								B := I;
							end;
						P := P^.Next;
						Inc (I);
					end;
				if N = 0 then
					Exit
				else
					if N = 1 then
						begin
							SetView := 0;
							Byte (Data) := B;
							Exit;
						end;
			end;
	end;

procedure TCMobile.NextState (By : Boolean; Var Data);
	begin
		if Max > 0 then
			if By then
				begin
					Inc (Byte (Data));
					if Byte (Data) >= Max then Byte (Data) := 0
				end
			else
				begin
					Dec (Byte (Data));
					if Byte (Data) >= Max then Byte (Data) := Max - 1;
				end;
	end;

destructor TCMobile.Done;
	begin
		TColumn.Done;
		Base^.Free;
	end;

{************************************************************** TMobileRec **}
constructor TMobileRec.Init (AName : PString; ANext : PMobileRec);
	begin
		Name := AName;
		Next := ANext;
	end;

destructor TMobileRec.Done;
	begin
		Next^.Free;
		DisposeStr (Name);
	end;

{*************************************************************** TCLongint **}
constructor TCLongint.Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
	begin
		TColumn.Init (ATitle, SizeOf (Longint), 11, AHeight, nil, ANext);
		Flags := Flags or cfNumber;
	end;

procedure TCLongint.GetView (Var S : String; Var Data);
	begin
		Str (Longint (Data) : Width - 1, S);
	end;

function TCLongint.SetView (S : String; Var Data) : Integer;
	Var
		I : Integer;
	begin
		ForNumber (S);
		Val (S, Longint (Data), I);
		SetView := I;
	end;

{****************************************************************** TCByte **}
constructor TCByte.Init (ATitle : PString; AWidth, AHeight : Integer; ANext : PColumn);
	begin
		TColumn.Init (ATitle, SizeOf (Byte), AWidth, AHeight, nil, ANext);
		Flags := Flags or cfNumber;
	end;

procedure TCByte.GetView (Var S : String; Var Data);
	begin
		Str (Byte (Data) : Width - 1, S);
	end;

function TCByte.SetView (S : String; Var Data) : Integer;
	Var
		I : Integer;
	begin
		ForNumber (S);
		Val (S, Byte (Data), I);
		SetView := I;
	end;

{***************************************************************** TCYesNo **}
constructor TCYesNo.Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
	begin
		TColumn.Init (ATitle, SizeOf (Boolean), 4, AHeight, nil, ANext);
		Flags := Flags or cfMobile;
	end;

procedure TCYesNo.GetView (Var S : String; Var Data);
	Var
		I : Integer;
	begin
		if Boolean (Data) then
			S := ''
		else
			S := '';
		I := (Width - Length (S)) shr 1;
		S := Space (I) + S + Space (Width - I - Length (S));
	end;

function TCYesNo.SetView (S : String; Var Data) : Integer;
	begin
		SetView := 0;
		ForMobile (S);
		case S [1] of
			'', 'Y', 'T', '1' : Boolean (Data) := True;
			'H', 'N', 'F', '0',	' ' : Boolean (Data) := False;
		else
			SetView := 1;
		end;
	end;

procedure TCYesNo.NextState (By : Boolean; Var Data);
	begin
		Boolean (Data) := not Boolean (Data);
	end;

{*************************************************************** TCInteger **}
constructor TCInteger.Init (ATitle : PString; AHeight : Integer; ANext : PColumn);
	begin
		TColumn.Init (ATitle, SizeOf (Integer), 6, AHeight, nil, ANext);
		Flags := Flags or cfNumber;
	end;

procedure TCInteger.GetView (Var S : String; Var Data);
	begin
		Str (Integer (Data) : Width - 1, S);
	end;

function TCInteger.SetView (S : String; Var Data) : Integer;
	Var
		I : Integer;
	begin
		ForNumber (S);
		Val (S, Integer (Data), I);
		SetView := I;
	end;

{**************************************************************** TCString **}
constructor TCString.Init (ATitle : PString; AWidth, AHeight : Integer; ANext : PColumn);
	begin
		if AWidth > 256 then AWidth := 256;
		TColumn.Init (ATitle, AWidth, AWidth, AHeight, nil, ANext);
	end;

procedure TCString.GetView (Var S : String; Var Data);
	begin
		CutSpace (S);
		S := String (Data);
	end;

function TCString.SetView (S : String; Var Data) : Integer;
	begin
		String (Data) := S;
		SetView := 0;
	end;

{***************************************************************** TCFloat **}
constructor TCFloat.Init (ATitle : PString; AWidth, ADecWid, AHeight : Integer; ANext : PColumn);
	begin
		TColumn.Init (ATitle, SizeOf (Real), AWidth, AHeight, nil, ANext);
		DecWid := ADecWid;
		Flags := Flags or cfNumber;
	end;

procedure TCFloat.GetView (Var S : String; Var Data);
	begin
		Str (Real (Data) : Width - 1 : DecWid, S);
		ReplaceChar (S, '.', ',');
	end;

function TCFloat.SetView (S : String; Var Data) : Integer;
	Var
		I : Integer;
	begin
		ForNumber (S);
		ReplaceChar (S, ',', '.');
		Val (S, Real (Data), I);
		SetView := I;
	end;

{**************************************************************** Routines **}
procedure ForNumber (Var S : String);
	begin
		RemoveSpace (S);
		AlfaUpStr (S);
		ReplaceChar (S, 'O', '0');
		if S = '' then S := '0';
	end;

procedure ForMobile (Var S : String);
	begin
		CutSpace (S);
		AlfaUpStr (S);
		if S = '' then S := ' ';
	end;
end.