{Copyright by Leonid Youriev, 1993, os@imfiko.bishkek.su}
{$I VERSION}

unit Grid;

interface

uses	Drivers, Objects, Views, Memory;

{$I VERSION}

type

	PLinker = ^TLinker;
	TLinker = Object (TView)
		Link		: PView;
		constructor Init (ByKey : Boolean; ALink : PView; AItems : Pointer);
		constructor	Load (Var S : TStream);
		procedure	Store (Var S : TStream);
		procedure	HandleEvent (Var Event : TEvent); virtual;
		destructor	Done; virtual;
	private
		Items		: Pointer;
	end;

	PGrid = ^TGrid;
	TGrid = Object (TView)
		constructor Init (Var By : TGroup);
		constructor	Load (Var S : TStream);
		procedure	Store (Var S : TStream);
		procedure	HandleEvent (Var Event : TEvent); virtual;
		function	FindCurrent : Longint;
		destructor	Done; virtual;
	private
		Items		: PPtrArray;
		X, Y		: Integer;
	end;

	PSoftGrid = ^TSoftGrid;
	TSoftGrid = Object (TView)
		constructor Init (Var By : TGroup);
		procedure	HandleEvent (Var Event : TEvent); virtual;
	end;

function MakeItems (Size : Word) : Pointer;

procedure GridItem (P : PView);
inline ($90);

procedure NewGridLine;
inline (
	$33/$C0/	{	XOR		AX, AX		}
	$50/		{	PUSH	AX			}
	$50			{	PUSH	AX			}
	);

procedure PureItem;
inline (
	$54/		{	PUSH	SP			}
	$33/$C0/	{	XOR		AX, AX		}
	$50			{	PUSH	AX			}
	);

procedure EndGrid;
inline (
	$33/$C0/	{	XOR		AX, AX		}
	$50/		{	PUSH	AX			}
	$50/		{	PUSH	AX			}
	$50/		{	PUSH	AX			}
	$50			{	PUSH	AX			}
	);

implementation

constructor TSoftGrid.Init (Var By : TGroup);
begin
	TObject.Init;
	EventMask := evKeyDown;
	Options := ofPostProcess;
	By.Insert (@Self);
end;

type
	WhatFind	= function (X1, Y1, X2, Y2 : Integer) : Word;

function Find (By : PGroup; What : WhatFind) : PView;
var
	X, Y	: Integer;
	N, P	: PView;
	Min, I	: Word;
begin
	Min := MaxInt;
	N := NIL;
	P := By^.Last;
	with By^.Current^ do
		begin
			X := Origin.X + Size.X shr 1;
			Y := Origin.Y + Size.Y shr 1;
		end;
	repeat
		with P^ do
			if (Options and ofSelectable <> 0) and (State and sfVisible <> 0) then
				begin
					I := What (X, Y, Origin.X + Size.X shr 1, Origin.Y + Size.Y shr 1);
					if I < Min then
						begin
							Min := I;
							N := P;
						end;
				end;
			P := P^.Next;
	until (PtrRec (P).Seg = 0) or (P = By^.Last);
	Find := N;
end;

function ToLeft (X1, Y1, X2, Y2 : Integer) : Word; far; assembler;
asm
	MOV		AX, MaxInt
	MOV		DX, X1
	SUB		DX, X2
	JLE		@2
	MOV		BX, Y1
	SUB		BX, Y2
	JG		@1
	NEG		BX
@1:	ADD		BX, BX
	ADD		BX, BX
	JC		@2
	ADD		DX, BX
	XCHG	AX, DX
@2:
end;

function ToRight (X1, Y1, X2, Y2 : Integer) : Word; far; assembler;
asm
	PUSH	X2
	PUSH	Y1
	PUSH	X1
	PUSH	Y2
	CALL	ToLeft
end;

function ToUp (X1, Y1, X2, Y2 : Integer) : Word; far; assembler;
asm
	PUSH	Y1
	PUSH	X1
	PUSH	Y2
	PUSH	X2
	CALL	ToLeft
end;

function ToDown (X1, Y1, X2, Y2 : Integer) : Word; far; assembler;
asm
	PUSH	Y2
	PUSH	X1
	PUSH	Y1
	PUSH	X2
	CALL	ToLeft
end;

procedure TSoftGrid.HandleEvent (Var Event : TEvent);
var
	P		: PView;
	As		: WhatFind;
	X1, Y1,
	X2, Y2,
	XC, YC,
	X, Y	: Integer;
label
	St;
begin
	if Owner^.Phase = phPostProcess then
		begin
			with Owner^.Current^ do
				begin
					X1 := Origin.X; Y1 := Origin.Y;
					X2 := X1 + Size.X; Y2 := Y1 + Size.Y;
					XC := (X1 + X2) shr 1; YC := (Y1 + Y2) shr 1;
					Dec (X1);
					Dec (Y1);
				end;
			case Event.KeyCode of
				kbUp	: As := ToUp;
				kbDown	: As := ToDown;
				kbLeft	: As := ToLeft;
				kbRight	: As := ToRight;
			else
				case Event.KeyCode of
					kbCtrlLeft,
					kbHome	:
						begin
							X := 1; Y := YC;
						end;
					kbCtrlRight,
					kbEnd	:
						begin
							X := Owner^.Size.X; Y := YC;
						end;
					kbCtrlUp,
					kbPgUp	:
						begin
							X := XC; Y := 1;
						end;
					kbCtrlDown,
					kbPgDn	:
						begin
							X := XC; Y := Owner^.Size.Y;
						end;
				else
					Exit;
				end;
				P := Owner^.Search (X, Y);
				goto St;
			end;
			P := Find (Owner, As);
			St:
			if (PtrRec (P).Seg <> 0) and P^.Focus then
				ClearEvent (Event);
		end;
end;

constructor TGrid.Init;
var
	LX, LY,
	I, N	: Integer;
label
	No;
begin
	TObject.Init;
	asm
		XOR		AX, AX
		MOV		LX, AX
		MOV		LY, AX
		MOV		SI, [BP]
		MOV		SI, SS:[SI + 2]
		STD
		MOV		BX, -4
		LODSW
		PUSH	SI
		PUSH	DS
		PUSH	SS
		POP		DS
		MOV		CX, -1
	@1:	INC		LY
	@2: ADD		BX, 4
		INC		CX
		LODSW
		XCHG	AX, DX
		LODSW
		OR		AX, DX
		JNE		@2
		CMP		CX, LX
		JB		@3
		MOV		LX, CX
	@3: XOR		CX, CX
		LODSW
		XCHG	AX, DX
		LODSW
		OR		AX, DX
		JNE		@1
		POP		DS
		PUSH	BX
		CALL	MemAlloc
		TEST	DX, DX
		POP		SI
		JE		No
		LES		DI, Self
		MOV		WORD PTR ES:[DI].Items, AX
		MOV		WORD PTR ES:[DI].Items [2], DX
		MOV		ES, DX
		MOV		DI, AX
		MOV		CH, BYTE PTR LY
	@4:	MOV		CL, BYTE PTR LX
	@5: STD
		SEGSS	LODSW
		XCHG	AX, DX
		SEGSS	LODSW
		CLD
		MOV		BX, AX
		OR		BX, DX
		JE		@6
		STOSW
		XCHG	AX, DX
		STOSW
		DEC		CL
		JNE		@5
		SUB		SI, 4
	@6:	PUSH	CX
		MOV		CH, 0
		ADD		CX, CX
		REP		STOSW
		POP		CX
		DEC		CH
		JNE		@4
	end;
	X := LX;
	Y := LY;
	EventMask := evKeyDown;
	Options := ofPostProcess;
	with By do
		begin
			N := LX * LY - 1;
			for I := 0 to N do
				Insert (PView (Items^ [I]));
			Insert (@Self);
		end;
	Exit;
	No:
	Fail;
end;

constructor TGrid.Load (Var S : TStream);
var
	I	: Integer;
begin
{$IFNDEF Turbo6}
	inherited Load (S);
{$ELSE}
	TView.Load (S);
{$ENDIF}
	S.Read (X, SizeOf (X) + SizeOf (Y));
	Items := MemAlloc (SizeOf (Pointer) * X * Y);
	if PtrRec (Items).Seg <> 0 then
		for I := X * Y - 1 downto 0 do
			GetPeerViewPtr (S, Items^ [I])
	else
		S.Status := stNoMemory;
end;

procedure TGrid.Store (Var S : TStream);
var
	I	: Integer;
begin
{$IFNDEF Turbo6}
	inherited Store (S);
{$ELSE}
	TView.Store (S);
{$ENDIF}
	S.Write (X, SizeOf (X) + SizeOf (Y));
	for I := X * Y - 1 downto 0 do
		PutPeerViewPtr (S, PView (Items^ [I]));
end;

function TGrid.FindCurrent : Longint; assembler;
asm
	LES		DI, Self
	XOR		AX, AX
	CMP		WORD PTR ES:[DI].Owner [2], AX
	JE		@4
	CMP		WORD PTR ES:[DI].Items [2], AX
	JE		@4
	MOV		AX, ES:[DI].Y
	MUL		ES:[DI].X
	XCHG	AX, CX
	LES		DI, ES:[DI].Owner
	MOV		AX, WORD PTR ES:[DI].TGroup.Current
	MOV		DX, WORD PTR ES:[DI].TGroup.Current [2]
	TEST	DX, DX
	JE		@4
	LES		DI, Self
	LES		DI, ES:[DI].Items
	MOV		BX, CX
	CLD
@1: SCASW
	JE		@3
@2:	ADD		DI, 2
	LOOP	@1
	JMP		@4
@3:	CMP		DX, ES:[DI]
	JNE		@2
	SUB		BX, CX
	LES		DI, Self
	XCHG	AX, BX
	CWD
	DIV		ES:[DI].X
	JMP		@5
@4:	MOV		AX, -1
@5:
end;

procedure TGrid.HandleEvent (Var Event : TEvent);
var
	T	: record
		case Integer of
			1: (Y, X	: Integer);
			0: (L		: Longint);
		end;
	D	: Integer;

	function Delta (N : Integer) : Boolean;
	begin
		if D >= 0 then
			D := -(D + 1)
		else
			D := -D;
		Delta := abs (D) > N shr 1;
	end;

	function ToView (NX, NY : Integer) : Boolean;
	var
		P	: PView;
	begin
		ToView := False;
		P := Items^ [NY * X + NX];
		if PtrRec (P).Seg <> 0 then
			begin
				P^.Focus;
				ToView := True;
			end;
	end;

label
	St;
begin
	if Owner^.Phase <> phPostProcess then Exit;
	T.L := FindCurrent;
	if T.Y >= 0 then
		begin
			D := 0;
			case Event.KeyCode of
				kbUp	:
					repeat
						while T.Y > 0 do
							begin
								Dec (T.Y);
								if ToView (T.X + D, T.Y) then goto St
							end;
					until Delta (T.X);
				kbDown	:
					repeat
						while T.Y < Y - 1 do
							begin
								Inc (T.Y);
								if ToView (T.X + D, T.Y) then goto St
							end;
					until Delta (T.X);
				kbLeft	:
					repeat
						while T.X > 0 do
							begin
								Dec (T.X);
								if ToView (T.X, T.Y + D) then goto St
							end;
					until Delta (T.Y);
				kbRight	:
					repeat
						while T.X < X - 1 do
							begin
								Inc (T.X);
								if ToView (T.X, T.Y + D) then goto St
							end;
					until Delta (T.Y);
				kbCtrlUp,
				kbHome	:
					for T.X := 0 to X - 1 do
						if ToView (T.X, T.Y) then goto St;
				kbCtrlRight,
				kbEnd	:
					for T.X := X - 1 downto 0 do
						if ToView (T.X, T.Y) then goto St;
				kbCtrlUp,
				kbPgUp	:
					for T.Y := 0 to Y - 1 do
						if ToView (T.X, T.Y) then goto St;
				kbCtrlDown,
				kbPgDn		:
					for T.Y := Y - 1 downto 0 do
						if ToView (T.X, T.Y) then goto St;
				kbCtrlPgUp,
				kbCtrlHome	:
					for T.Y := 0 to Y - 1 do
						for T.X := 0 to X - 1 do
							if ToView (T.X, T.Y) then goto St;
				kbCtrlPgDn,
				kbCtrlEnd	:
					for T.Y := Y - 1 downto 0 do
						for T.X := X - 1 downto 0 do
							if ToView (T.X, T.Y) then goto St;
			end;
			Exit;
			St:
			ClearEvent (Event);
		end;
end;

destructor TGrid.Done;
begin
	if PtrRec (Items).Seg > 0 then
		FreeMem (Items, SizeOf (Pointer) * X * Y);
{$IFNDEF Turbo6}
	inherited Done;
{$ELSE}
	TView.Done;
{$ENDIF}
end;

type
	TLinkerItem = record
		KeyCode	: Word;
		case Integer of
			0	: (View		: PView);
			1	: (Message	: Word);
	end;

	PLinkerItemList = ^TLinkerItemList;
	TLinkerItemList = record
		Count	: Word;
		Items	: array [0..255] of TLinkerItem;
	end;

function MakeItems (Size : Word) : Pointer; assembler;
asm
	MOV		AX, Size
	ADD		AX, 2
	PUSH	AX
	CALL	MemAlloc
	MOV		DI, AX
	MOV		ES, DX
	MOV		SI, [BP]
	MOV		SI, SS:[SI + 2]
	STD
	LODSW
	TEST	DX, DX
	JE		@2
	MOV		AX, Size
	XOR		DX, DX
	MOV		CX, 6
	DIV		CX
	MOV		ES:[DI], AX
	XCHG	AX, CX
	JCXZ	@2
	MOV		BX, CX
	ADD		BX, BX
	ADD		DI, BX
	ADD		BX, BX
	ADD		DI, BX
	NEG		BX
@1:	SEGSS	MOVSW
	SEGSS	MOVSW
	SEGSS	LODSW
	ADD		BX, 2
	MOV		ES:[DI + BX + 2], AX
	LOOP	@1
@2:	LEA		AX, DI [BX]
	MOV		DX, ES
end;

constructor TLinker.Init (ByKey : Boolean; ALink : PView; AItems : Pointer);
begin
	TObject.Init;
	Link := ALink;
	Items := AItems;
	Options := ofPostProcess;
	if ByKey then
		EventMask := evKeyDown
	else
		EventMask := evCommand;
end;

constructor TLinker.Load (Var S : TStream);
var
	C	: Word;
begin
{$IFNDEF Turbo6}
	inherited Load (S);
{$ELSE}
	TView.Load (S);
{$ENDIF}
	GetPeerViewPtr (S, Link);
	S.Read (C, SizeOf (C));
	Items := MemAlloc (SizeOf (C) + C * SizeOf (TLinkerItem));
	if PtrRec (Items).Seg <> 0 then
		with TLinkerItemList (Items^) do
			begin
				Count := C;
				S.Read (Items, C * SizeOf (TLinkerItem));
			end
	else
		S.Status := stNoMemory;
end;

procedure TLinker.Store (Var S : TStream);
begin
{$IFNDEF Turbo6}
	inherited Store (S);
{$ELSE}
	TView.Store (S);
{$ENDIF}
	PutPeerViewPtr (S, Link);
	with TLinkerItemList (Items^) do
		S.Write (Count, SizeOf (Count) + Count * SizeOf (TLinkerItem));
end;

procedure TLinker.HandleEvent (Var Event : TEvent);
label
	Sx, No;
begin
	if (Owner^.Phase = phPostProcess) and (PtrRec (Link).Seg <> 0) and (Link^.State and sfActive <> 0) then
		begin
			asm
				LES		DI, Event
				MOV		DX, ES:[DI].TEvent.KeyCode
				LES		SI, Self
				LES		SI, ES:[SI].Items
				MOV		CX, ES
				JCXZ	No
				CLD
				SEGES	LODSW
				XCHG	AX, CX
				JCXZ	No
				MOV		DI, SI
				XCHG	AX, DX
				REPNE	SCASW
				JNE		No
				SUB		DI, SI
				ADD		SI, DI
				ADD		SI, DI
				ADD		SI, DI
				STD
				LODSW
				SEGES	LODSW
				XCHG	AX, CX
				SEGES	LODSW
				JCXZ	@1
				PUSH	CX
				PUSH	AX
				CALL	Focus
				OR		AL, AL
				JE		No
				JMP		Sx
			@1: TEST	AX, AX
				JE		Sx
				PUSH	WORD PTR Self [2]
				PUSH	WORD PTR Self
				PUSH	AX
				CALL	SendCommand
			end;
			Sx:
			ClearEvent (Event);
			No:
		end;
end;

destructor TLinker.Done;
begin
	if PtrRec (Items).Seg > 0 then
		FreeMem (Items, SizeOf (Word) + Word (TLinkerItemList (Items^).Count) * SizeOf (TLinkerItem));
{$IFNDEF Turbo6}
	inherited Done;
{$ELSE}
	TView.Done;
{$ENDIF}
end;

end.