{1/2 Copyright by Leonid Youriev 1992, 93}
{$I VERSION}

unit Menus;

interface

uses	Objects, Drivers, Views, Styles;

{$I VERSION}

type

	TMenuStr = String [31];
	HintFunc = function (AHelpCtx : Word) : String;

	PMenu = ^TMenu;

	PMenuItem = ^TMenuItem;
	TMenuItem = Record
		Next		: PMenuItem;
		Name		: PString;
		Command		: Word;
		Disabled	: Boolean;
		KeyCode		: Word;
		HelpCtx		: Word;
		case Integer of
			0	: (Param	: PString);
			1	: (SubMenu	: PMenu);
		end;

	TMenu = Record
		Items	: PMenuItem;
		Default	: PMenuItem;
	end;

{ TMenuView object }
	{ Palette layout }
	{ 1 = Normal text }
	{ 2 = Disabled text }
	{ 3 = Shortcut text }
	{ 4 = Normal selection }
	{ 5 = Disabled selection }
	{ 6 = Shortcut selection }

	PMenuView = ^TMenuView;
	TMenuView = Object (TView)
		ParentMenu	: PMenuView;
		Menu		: PMenu;
		Current		: PMenuItem;
		constructor	Init (var Bounds : TRect);
		constructor	Load (var S : TStream);
		function	Execute : Word; virtual;
		function	FindItem (Ch : Char) : PMenuItem;
		procedure	GetItemRect (Item : PMenuItem; var R : TRect); virtual;
		function	GetHelpCtx : Word; virtual;
		procedure	HandleEvent (var Event : TEvent); virtual;
		function	HotKeyOfItem (KeyCode : Word) : PMenuItem;
		function	NewSubView (var Bounds : TRect; AMenu : PMenu; AParentMenu : PMenuView) : PMenuView; virtual;
		procedure	Store (var S : TStream);
	end;

{ TMenuBar object }
	{ Palette layout }
	{ 1 = Normal text }
	{ 2 = Disabled text }
	{ 3 = Shortcut text }
	{ 4 = Normal selection }
	{ 5 = Disabled selection }
	{ 6 = Shortcut selection }

	PMenuBar = ^TMenuBar;
	TMenuBar = Object (TMenuView)
		constructor	Init (var Bounds : TRect; AMenu : PMenu);
		destructor	Done; virtual;
		procedure	Draw; virtual;
		procedure	GetItemRect (Item : PMenuItem; var R : TRect); virtual;
	end;

{ TMenuBox object }
	{ Palette layout }
	{ 1 = Normal text }
	{ 2 = Disabled text }
	{ 3 = Shortcut text }
	{ 4 = Normal selection }
	{ 5 = Disabled selection }
	{ 6 = Shortcut selection }

	PMenuBox = ^TMenuBox;
	TMenuBox = object(TMenuView)
		constructor	Init (var Bounds : TRect; AMenu : PMenu; AParentMenu : PMenuView);
		procedure	Draw; virtual;
		procedure	GetItemRect (Item : PMenuItem; var R : TRect); virtual;
	end;

{ TStatusItem }
	PStatusItem = ^TStatusItem;
	TStatusItem = Record
		Next	: PStatusItem;
		Text	: PString;
		KeyCode	: Word;
		Command	: Word;
	end;

{ TStatusDef }
	PStatusDef = ^TStatusDef;
	TStatusDef = Record
		Next	: PStatusDef;
		Min,
		Max		: Word;
		Items	: PStatusItem;
	end;

{ TStatusLine }
	{ Palette layout }
	{ 1 = Normal text }
	{ 2 = Disabled text }
	{ 3 = Shortcut text }
	{ 4 = Normal selection }
	{ 5 = Disabled selection }
	{ 6 = Shortcut selection }

	PStatusLine = ^TStatusLine;
	TStatusLine = object(TView)
		Items		: PStatusItem;
		Defs		: PStatusDef;
		constructor	Init (var Bounds : TRect; ADefs : PStatusDef);
		constructor	Load (var S : TStream);
		destructor	Done; virtual;
		procedure	Draw; virtual;
		procedure	HandleEvent (var Event : TEvent); virtual;
		procedure	Store (var S : TStream);
		procedure	Update; virtual;
	private
		procedure	DrawSelect (Selected : PStatusItem);
		procedure	FindItems;
	end;

{ TMenuItem routines }
	function NewItem (AName, AParam : TMenuStr; AKeyCode : Word; ACommand : Word; AHelpCtx : Word; ANext : PMenuItem) : PMenuItem;
	function NewLine (ANext : PMenuItem) : PMenuItem;
	function NewSubMenu (AName : TMenuStr; AHelpCtx : Word; ASubMenu : PMenu; ANext : PMenuItem) : PMenuItem;

{ TMenu routines }
	function NewMenu (Items : PMenuItem) : PMenu;
	procedure DisposeMenu (Menu : PMenu);

{ TStatusLine routines }
	function NewStatusDef (AMin, AMax : Word; AItems : PStatusItem; ANext : PStatusDef) : PStatusDef;
	function NewStatusKey (AText : String; AKeyCode : Word; ACommand : Word; ANext : PStatusItem) : PStatusItem;

{ Menus registration procedure }
	procedure RegisterMenus;

const

	Hint			: HintFunc = NIL;
	MenuFrameStyle	: TMenuFrameStyles = msDefault;

{********************************************* Stream registration records **}
RMenuBar	: TStreamRec = (
	ObjType	: 40;
	VmtLink	: Ofs (TypeOf (TMenuBar)^);
	Load	: @TMenuBar.Load;
	Store	: @TMenuBar.Store);

RMenuBox	: TStreamRec = (
	ObjType	: 41;
	VmtLink	: Ofs (TypeOf (TMenuBox)^);
	Load	: @TMenuBox.Load;
	Store	: @TMenuBox.Store);

RStatusLine	: TStreamRec = (
	ObjType	: 42;
	VmtLink	: Ofs (TypeOf (TStatusLine)^);
	Load	: @TStatusLine.Load;
	Store	: @TStatusLine.Store);

implementation

uses	CharLog;

{****************************************************** TMenuItem routines **}
function NewItem (AName, AParam : TMenuStr; AKeyCode : Word; ACommand : Word; AHelpCtx : Word; ANext : PMenuItem) : PMenuItem;
var
	P	: PMenuItem;
	T	: PView;
begin
	if (AName <> '') and (ACommand <> 0) then
		begin
			New (P);
			with P^ do
				begin
					Next := ANext;
					Name := GenStr (@AName);
					Command := ACommand;
					Disabled := not T^.CommandEnabled (ACommand);
					KeyCode := AKeyCode;
					HelpCtx := AHelpCtx;
					Param := GenStr (@AParam);
				end;
			NewItem := P;
		end
	else
		NewItem := ANext;
end;

function NewLine (ANext : PMenuItem) : PMenuItem;
var
	P	: PMenuItem;
begin
	New (P);
	with P^ do
		begin
			Next := ANext;
			Name := NIL;
			HelpCtx := hcNoContext;
		end;
	NewLine := P;
end;

function NewSubMenu (AName : TMenuStr; AHelpCtx: Word; ASubMenu : PMenu; ANext : PMenuItem) : PMenuItem;
var
	P	: PMenuItem;
begin
	if (AName <> '') and (PtrRec (ASubMenu).Seg <> 0) then
		begin
			New (P);
			with P^ do
				begin
					Next := ANext;
					Name := GenStr (@AName);
					Command := 0;
					Disabled := False;
					HelpCtx := AHelpCtx;
					SubMenu := ASubMenu;
				end;
			NewSubMenu := P;
		end
	else
		NewSubMenu := ANext;
end;

{********************************************************** TMenu routines **}
function NewMenu (Items : PMenuItem) : PMenu;
var
	P	: PMenu;
begin
	New (P);
	P^.Items := Items;
	P^.Default := Items;
	NewMenu := P;
end;

procedure DisposeMenu (Menu : PMenu);
var
	P,
	Q	: PMenuItem;
begin
	if PtrRec (Menu).Seg <> 0 then
		begin
			P := Menu^.Items;
			while PtrRec (P).Seg <> 0 do
				begin
					if PtrRec (P^.Name).Seg <> 0 then
						begin
							DisposeStr (P^.Name);
							if P^.Command <> 0 then
								DisposeStr (P^.Param)
							else
								DisposeMenu (P^.SubMenu);
						end;
					Q := P;
					P := P^.Next;
					Dispose (Q);
				end;
		Dispose (Menu);
	end;
end;

{*************************************************************** TMenuView **}
constructor TMenuView.Init (var Bounds : TRect);
begin
	TView.Init (Bounds);
	EventMask := EventMask or evBroadcast;
end;

constructor TMenuView.Load (var S : TStream);

function DoLoadMenu : PMenu;
	var
		Item	: PMenuItem;
		Last	: ^PMenuItem;
		Menu	: PMenu;
		Tok		: Byte;
	begin
		New (Menu);
		Last := @Menu^.Items;
		Item := NIL;
		S.Read (Tok, 1);
		while Tok <> 0 do
			begin
				New (Item);
				Last^ := Item;
				Last := @Item^.Next;
				with Item^ do
					begin
						Name := S.ReadStr;
						S.Read (Command, SizeOf (Word) * 3 + SizeOf (Boolean));
						if PtrRec (Name).Seg <> 0 then
							if Command = 0 then
								SubMenu := DoLoadMenu
							else
								Param := S.ReadStr;
					end;
				S.Read (Tok, 1);
			end;
		Last^ := NIL;
		Menu^.Default := Menu^.Items;
		DoLoadMenu := Menu;
	end;

begin
	TView.Load (S);
	Menu := DoLoadMenu;
end;

function TMenuView.Execute : Word;
type
	MenuAction = (DoNothing, DoSelect, DoReturn);
var
	AutoSelect	: Boolean;
	Action		: MenuAction;
	Ch			: Char;
	Result		: Word;
	ItemShown,
	P			: PMenuItem;
	Target		: PMenuView;
	R			: TRect;
	E			: TEvent;

procedure TrackMouse;
	var
		Mouse	: TPoint;
		R		: TRect;
	begin
		MakeLocal (E.Where, Mouse);
		Current := Menu^.Items;
		while PtrRec (Current).Seg <> 0 do
			begin
				GetItemRect (Current, R);
				if R.Contains (Mouse) then Exit;
				Current := Current^.Next;
			end;
	end;

procedure TrackKey (FindNext : Boolean);

	procedure NextItem;
		begin
			Current := Current^.Next;
			if PtrRec (Current).Seg = 0 then Current := Menu^.Items;
		end;

	procedure PrevItem;
		var
			P	: PMenuItem;
		begin
			P := Current;
			if P = Menu^.Items then P := NIL;
			repeat NextItem until Current^.Next = P;
		end;

begin
	if PtrRec (Current).Seg <> 0 then
		repeat
			if FindNext then
				NextItem
			else
				PrevItem;
		until PtrRec (Current^.Name).Seg <> 0;
end;

function MouseInOwner : Boolean;
	var
		Mouse	: TPoint;
		R		: TRect;
	begin
		MouseInOwner := False;
		if (PtrRec (ParentMenu).Seg <> 0) and (ParentMenu^.Size.Y = 1) then
			begin
				ParentMenu^.MakeLocal (E.Where, Mouse);
				ParentMenu^.GetItemRect (ParentMenu^.Current, R);
				MouseInOwner := R.Contains (Mouse);
			end;
	end;

function MouseInMenus : Boolean;
	var
		P	: PMenuView;
	begin
		P := ParentMenu;
		while (PtrRec (P).Seg <> 0) and not P^.MouseInView (E.Where) do
			P := P^.ParentMenu;
		MouseInMenus := PtrRec (P).Seg <> 0;
	end;

function TopMenu : PMenuView;
	var
		P	: PMenuView;
	begin
		P := @Self;
		while PtrRec (P^.ParentMenu).Seg <> 0 do
			P := P^.ParentMenu;
		TopMenu := P;
	end;

begin
	AutoSelect := False;
	Result := 0;
	ItemShown := NIL;
	Current := Menu^.Default;
	repeat
		Action := DoNothing;
		GetEvent (E);
		case E.What of
			evMouseDown	:
				if MouseInView (E.Where) or MouseInOwner then
					begin
						TrackMouse;
						if Size.Y = 1 then AutoSelect := True;
					end
				else
					Action := DoReturn;
			evMouseUp	:
				begin
					TrackMouse;
					if MouseInOwner then
						Current := Menu^.Default
					else
						if (PtrRec (Current).Seg <> 0) and (PtrRec (Current^.Name).Seg <> 0) then
							Action := DoSelect
						else
							Action := DoReturn;
				end;
			evMouseMove	:
				if E.Buttons <> 0 then
					begin
						TrackMouse;
						if not (MouseInView (E.Where) or MouseInOwner) and MouseInMenus then
							Action := DoReturn;
					end;
			evKeyDown	:
				case CtrlToArrow (E.KeyCode) of
					kbUp,
					kbDown	:
						if Size.Y <> 1 then
							TrackKey (CtrlToArrow (E.KeyCode) = kbDown)
						else
							if E.KeyCode = kbDown then AutoSelect := True;
					kbLeft,
					kbRight	:
						if PtrRec (ParentMenu).Seg = 0 then
							TrackKey (CtrlToArrow (E.KeyCode) = kbRight)
						else
							Action := DoReturn;
					kbHome,
					kbEnd	:
						if Size.Y <> 1 then
							begin
								Current := Menu^.Items;
								if E.KeyCode = kbEnd then TrackKey (False);
							end;
					kbEnter	:
						begin
							if Size.Y = 1 then AutoSelect := True;
							Action := DoSelect;
						end;
					kbEsc	:
						begin
							Action := DoReturn;
							if (PtrRec (ParentMenu).Seg = 0) or (ParentMenu^.Size.Y <> 1) then
								ClearEvent (E);
						end;
				else
					Target := @Self;
					Ch := GetAltChar (E.KeyCode);
					if Ch = #0 then
						Ch := E.CharCode
					else
						Target := TopMenu;
					P := Target^.FindItem (Ch);
					if PtrRec (P).Seg = 0 then
						begin
							P := TopMenu^.HotKeyOfItem (E.KeyCode);
							if (PtrRec (P).Seg <> 0) and CommandEnabled (P^.Command) then
								begin
									Result := P^.Command;
									Action := DoReturn;
								end
						end
					else
						if Target = @Self then
							begin
								if Size.Y = 1 then AutoSelect := True;
								Action := DoSelect;
								Current := P;
							end
						else
							if (ParentMenu <> Target) or (ParentMenu^.Current <> P) then
								Action := DoReturn;
				end;
			evCommand	:
				if E.Command = cmMenu then
					begin
						AutoSelect := False;
						if PtrRec (ParentMenu).Seg <> 0 then Action := DoReturn;
					end
				else
					Action := DoReturn;
		end;
		if ItemShown <> Current then
			begin
				ItemShown := Current;
				DrawView;
			end;
		if ((Action = DoSelect) or ((Action = DoNothing) and AutoSelect)) and (PtrRec (Current).Seg <> 0) then
			with Current^ do
				if PtrRec (Name).Seg <> 0 then
					if Command = 0 then
						begin
							if E.What and (evMouseDown + evMouseMove) <> 0 then
								PutEvent (E);
							{  ! 㦥   ᤥ  ?}
							GetItemRect (Current, R);
							Inc (R.A.X, Origin.X);
							R.A.Y := R.B.Y + Origin.Y;
							R.B := Owner^.Size;
							if Size.Y = 1 then Dec (R.A.X);
							Target := TopMenu^.NewSubView (R, SubMenu, @Self);
							Result := Owner^.ExecView (Target);
							Dispose (Target, Done);
						end
					else
						if Action = DoSelect then Result := Command;
		if (Result <> 0) and CommandEnabled (Result) then
			begin
				Action := DoReturn;
				ClearEvent (E);
			end;
	until Action = DoReturn;
	if E.What <> evNothing then
		if (PtrRec (ParentMenu).Seg <> 0) or (E.What = evCommand) then PutEvent (E);
	if PtrRec (Current).Seg <> 0 then
		begin
			Menu^.Default := Current;
			Current	:= NIL;
			DrawView;
		end;
	Execute := Result;
end;

function TMenuView.FindItem (Ch : Char) : PMenuItem;
var
	P	: PMenuItem;
	I	: Integer;
begin
	Ch := KeyCase (Ch);
	P := Menu^.Items;
	while (Ch <> #0) and (PtrRec (P).Seg <> 0) do
		begin
			if (PtrRec (P^.Name).Seg <> 0) and not P^.Disabled then
				begin
					if Ch = KeyCase (HotKey (P^.Name^)) then
						begin
							FindItem := P;
							Exit;
						end;
				end;
			P := P^.Next;
		end;
	FindItem := NIL;
end;

procedure TMenuView.GetItemRect (Item : PMenuItem; var R : TRect);
begin
end;

function TMenuView.GetHelpCtx : Word;
var
	C	: PMenuView;
begin
	C := @Self;
	while (PtrRec (C).Seg <> 0) and ((PtrRec (C^.Current).Seg = 0)
	or (C^.Current^.HelpCtx = hcNoContext) or (PtrRec (C^.Current^.Name).Seg = 0)) do
		C := C^.ParentMenu;
	if PtrRec (C).Seg <> 0 then
		GetHelpCtx := C^.Current^.HelpCtx
	else
		GetHelpCtx := hcNoContext;
end;

procedure TMenuView.HandleEvent (var Event : TEvent);
var
	CallDraw	: Boolean;
	P			: PMenuItem;

procedure UpdateMenu (Menu : PMenu);
	var
		P				: PMenuItem;
		CommandState	: Boolean;
	begin
		P := Menu^.Items;
		while PtrRec (P).Seg <> 0 do
			begin
				if PtrRec (P^.Name).Seg <> 0 then
					if P^.Command = 0 then UpdateMenu (P^.SubMenu)
				else
					begin
						CommandState := CommandEnabled (P^.Command);
						if P^.Disabled = CommandState then
							begin
								P^.Disabled := not CommandState;
								CallDraw := True;
							end;
					end;
				P := P^.Next;
			end;
	end;

procedure DoSelect;
	begin
		PutEvent (Event);
		SendCommand (Owner^.ExecView (@Self));
		ClearEvent (Event);
	end;

begin
	if PtrRec (Menu).Seg <> 0 then
		case Event.What of
			evMouseDown	:	DoSelect;
			evKeyDown	:
				if (Seg (FindItem (GetAltChar (Event.KeyCode))^) <> 0) then
					DoSelect
				else
					begin
						P := HotKeyOfItem (Event.KeyCode);
						if (PtrRec (P).Seg <> 0) and SendCommand (P^.Command) then
							ClearEvent (Event);
					end;
			evCommand	:
				if Event.Command = cmMenu then DoSelect;
			evBroadcast	:
				if Event.Command = cmCommandSetChanged then
					begin
						CallDraw := False;
						UpdateMenu (Menu);
						if CallDraw then DrawView;
					end;
		end;
end;

function TMenuView.HotKeyOfItem (KeyCode : Word) : PMenuItem;

function FindHotKey (P : PMenuItem) : PMenuItem;
	var
		T	: PMenuItem;
	begin
		while PtrRec (P).Seg <> 0 do
			begin
				if PtrRec (P^.Name).Seg <> 0 then
					if P^.Command = 0 then
						begin
							T := FindHotKey (P^.SubMenu^.Items);
							if PtrRec (T).Seg <> 0 then
								begin
									FindHotKey := T;
									Exit;
								end;
						end
					else
						if not P^.Disabled and (P^.KeyCode <> kbNoKey) and (P^.KeyCode = KeyCode) then
							begin
								FindHotKey := P;
								Exit;
							end;
				P := P^.Next;
			end;
		FindHotKey := NIL;
	end;

begin
	HotKeyOfItem := FindHotKey (Menu^.Items);
end;

function TMenuView.NewSubView (var Bounds : TRect; AMenu : PMenu; AParentMenu : PMenuView) : PMenuView;
begin
	NewSubView := New (PMenuBox, Init (Bounds, AMenu, AParentMenu));
end;

procedure TMenuView.Store (var S : TStream);

procedure DoStoreMenu (Menu : PMenu);
	var
		Item	: PMenuItem;
		Tok		: Byte;
	begin
		Tok := $FF;
		Item := Menu^.Items;
		while PtrRec (Item).Seg <> 0 do
			begin
				with Item^ do
					begin
						S.Write (Tok, 1);
						S.WriteStr (Name);
						S.Write (Command, SizeOf (Word) * 3 + SizeOf (Boolean));
						if (PtrRec (Name).Seg <> 0) then
							if Command = 0 then
								DoStoreMenu (SubMenu)
							else
								S.WriteStr (Param);
					end;
				Item := Item^.Next;
			end;
		Tok := 0;
		S.Write (Tok, 1);
	end;

begin
	TView.Store (S);
	DoStoreMenu (Menu);
end;

{**************************************************************** TMenuBar **}
constructor TMenuBar.Init (var Bounds : TRect; AMenu : PMenu);
var
	R	: TRect;
begin
	TMenuView.Init (Bounds);
	GrowMode := gfGrowHiX;
	Menu := AMenu;
	Options := Options or ofPreProcess;
end;

destructor TMenuBar.Done;
begin
	TMenuView.Done;
	DisposeMenu (Menu);
end;

procedure TMenuBar.Draw;
var
	X, L	: Integer;
	CNormal,
	CSelect,
	CNormDisabled,
	CSelDisabled,
	Color	: Word;
	P		: PMenuItem;
	B		: TDrawBuffer;
	SCOff	: TSpecialCharsUser;

begin
	CNormal := GetColor ($0D0B);
	CSelect := GetColor ($100E);
	CNormDisabled := GetColor ($0C0C);
	CSelDisabled := GetColor ($0F0F);
	MoveChar (B, ' ', Lo (CNormal), Size.X);
	if PtrRec (Menu).Seg <> 0 then
		begin
			X := 1;
			P := Menu^.Items;
			while PtrRec (P).Seg <> 0 do
				begin
					if PtrRec (P^.Name).Seg <> 0 then
						begin
							L := CStrLen (P^.Name^);
							if P = Current then
								SCOff := scuMenuBar
							else
								SCOff := scuNone;
							if X + L < Size.X then
								begin
									if P^.Disabled then
										if P = Current then
											Color := CSelDisabled
										else
											Color := CNormDisabled
									else
										if P = Current then
											Color := CSelect
										else
											Color := CNormal;
									MoveMarkCStr (B [X], P^.Name^, Color, L + 2, SCOff);
								end;
							Inc (X, L + 2);
						end;
					P := P^.Next;
				end;
		end;
	WriteBuf (0, 0, Size.X, 1, B);
end;

procedure TMenuBar.GetItemRect (Item : PMenuItem; var R : TRect);
var
	P	: PMenuItem;
begin
	R.Assign (1, 0, 1, 1);
	P := Menu^.Items;
	while True do
		begin
			R.A.X := R.B.X;
			if PtrRec (P^.Name).Seg <> 0 then Inc (R.B.X, CStrLen (P^.Name^) + 2);
			if P = Item then Exit;
			P := P^.Next;
		end;
end;

{**************************************************************** TMenuBox **}
constructor TMenuBox.Init (var Bounds : TRect; AMenu : PMenu; AParentMenu : PMenuView);
var
	W,
	H,
	L	: Integer;
	P	: PMenuItem;
	R	: TRect;
begin
	W := 10;
	H := 2;
	if PtrRec (AMenu).Seg <> 0 then
		begin
			P := AMenu^.Items;
			while PtrRec (P).Seg <> 0 do
				begin
					if PtrRec (P^.Name).Seg <> 0 then
						begin
							L := CStrLen (P^.Name^) + 6;
							if P^.Command = 0 then Inc (L, 3) else
							if PtrRec (P^.Param).Seg <> 0 then Inc (L, CStrLen (P^.Param^) + 2);
							if L > W then W := L;
						end;
					Inc (H);
					P := P^.Next;
				end;
		end;
	R.Copy (Bounds);
	With R do
		begin
			if A.X + W < B.X then B.X := A.X + W else A.X := B.X - W;
			if A.Y + H < B.Y then B.Y := A.Y + H else A.Y := B.Y - H;
		end;
	TMenuView.Init (R);
	State := State or sfShadow;
	Options := Options or ofPreProcess;
	Menu := AMenu;
	ParentMenu := AParentMenu;
end;

procedure TMenuBox.Draw;
var
	CNormal,
	CSelect,
	CNormDisabled,
	CSelDisabled,
	Color			: Word;
	Y, DX, NX		: Integer;
	P				: PMenuItem;
	B				: TDrawBuffer;
	SCOff			: TSpecialCharsUser;

procedure FrameLine (N : Integer);
	begin
		MoveBuf (B [Size.X - 2], MenuFrameChars [N + 3], Lo (CNormal), 2);
		MoveBuf (B [0], MenuFrameChars [N], Lo (CNormal), 2);
		MoveChar (B [NX], MenuFrameChars [N + 2], Lo (Color), DX - NX + 1);
	end;

procedure DrawLine;
	begin
		WriteBuf (0, Y, Size.X, 1, B);
		Inc (Y);
	end;

begin
	CNormal := GetColor ($0D0B);
	CSelect := GetColor ($100E);
	CNormDisabled := GetColor ($0C0C);
	CSelDisabled := GetColor ($0F0F);
	Y := 0;
	Color := CNormal;
	if MenuFrameStyle >= msDefault then
		NX := 2
	else
		NX := 1;
	DX := Size.X - NX - 1;
	FrameLine (0);
	DrawLine;
	if PtrRec (Menu).Seg <> 0 then
		begin
			P := Menu^.Items;
			while PtrRec (P).Seg <> 0 do
				begin
					Color := CNormal;
					if PtrRec (P^.Name).Seg = 0 then
						FrameLine (15)
					else
						begin
							if P = Current then
								SCOff := scuMenuBox
							else
								SCOff := scuNone;
							if P^.Disabled then
								if P = Current then
									Color := CSelDisabled
								else
									Color := CNormDisabled
							else
								if P = Current then
									Color := CSelect;
							FrameLine (10);
							MoveCStr (B [NX + 1], P^.Name^, Color);
							if P^.Command = 0 then
								MoveChar (B [DX], '', Lo (Color), 1)
							else
								if PtrRec (P^.Param).Seg <> 0 then
									MoveStr (B [DX - Length (P^.Param^)], P^.Param^, Lo (Color));
							WordRec (B [NX]).Lo := Ord (SpecialChars [SCOff, 0]);
							WordRec (B [DX]).Lo := Ord (SpecialChars [SCOff, 1]);
						end;
					DrawLine;
					P := P^.Next;
				end;
		end;
	Color := CNormal;
	FrameLine (5);
	DrawLine;
end;

procedure TMenuBox.GetItemRect (Item : PMenuItem; var R : TRect);
var
	Y	: Integer;
	P	: PMenuItem;
begin
	Y := 1;
	P := Menu^.Items;
	while P <> Item do
		begin
			Inc (Y);
			P := P^.Next;
		end;
	if MenuFrameStyle >= msDefault then
		R.Assign (2, Y, Size.X - 2, Y + 1)
	else
		R.Assign (1, Y, Size.X - 1, Y + 1);
end;

{************************************************************* TStatusLine **}
constructor TStatusLine.Init (var Bounds : TRect; ADefs : PStatusDef);
begin
	TView.Init (Bounds);
	Options := Options or ofPreProcess;
	EventMask := EventMask or evBroadcast;
	GrowMode := gfGrowLoY + gfGrowHiX + gfGrowHiY;
	Defs := ADefs;
	FindItems;
end;

constructor TStatusLine.Load (var S : TStream);

function DoLoadStatusItems : PStatusItem;
	var
		Count		: Integer;
		Cur, First	: PStatusItem;
		Last		: ^PStatusItem;
	begin
		Cur := NIL;
		Last := @First;
		S.Read (Count, SizeOf (Integer));
		while Count > 0 do
			begin
				New (Cur);
				Last^ := Cur;
				Last := @Cur^.Next;
				Cur^.Text := S.ReadStr;
				S.Read (Cur^.KeyCode, SizeOf (Word) * 2);
				Dec (Count);
			end;
		Last^ := NIL;
		DoLoadStatusItems := First;
	end;

function DoLoadStatusDefs : PStatusDef;
	var
		AMin	: Word;
		Cur,
		First	: PStatusDef;
		Last	: ^PStatusDef;
		Count	: Integer;
	begin
		Last := @First;
		S.Read (Count, SizeOf (Integer));
		while Count > 0 do
			begin
				New (Cur);
				Last^ := Cur;
				Last := @Cur^.Next;
				S.Read (Cur^.Min, SizeOf (Word) shl 1);
				Cur^.Items := DoLoadStatusItems;
				Dec (Count);
			end;
		Last^ := NIL;
		DoLoadStatusDefs := First;
	end;

begin
	TView.Load (S);
	Defs := DoLoadStatusDefs;
	FindItems;
end;

destructor TStatusLine.Done;
var
	T	: PStatusDef;

procedure DisposeItems (Item : PStatusItem);
	var
		T	: PStatusItem;
	begin
		while PtrRec (Item).Seg <> 0 do
		begin
			T := Item;
			Item := Item^.Next;
			DisposeStr (T^.Text);
			Dispose (T);
		end;
	end;

begin
	while PtrRec (Defs).Seg <> 0 do
		begin
			T := Defs;
			Defs := Defs^.Next;
			DisposeItems (T^.Items);
			Dispose (T);
		end;
	TView.Done;
end;

procedure TStatusLine.Draw;
begin
	DrawSelect (NIL);
end;

procedure TStatusLine.DrawSelect (Selected : PStatusItem);
var
	B				: TDrawBuffer;
	T				: PStatusItem;
	I, L			: Integer;
	CSelect,
	CNormal,
	CSelDisabled,
	CNormDisabled	: Word;
	Color			: Word;
	HintBuf			: String;
	SCOff			: TSpecialCharsUser;

begin
	CNormal := GetColor ($0301);
	CSelect := GetColor ($0604);
	CNormDisabled := GetColor ($0202);
	CSelDisabled := GetColor ($0505);
	MoveChar (B, ' ', Lo (CNormal), Size.X);
	T := Items;
	I := 0;
	while PtrRec (T).Seg <> 0 do
		begin
			if PtrRec (T^.Text).Seg <> 0 then
				begin
					L := CStrLen (T^.Text^);
					if L > 0 then
						begin
							if I + L < Size.X then
								begin
									if T = Selected then
										SCOff := scuStatusLine
									else
										SCOff := scuNone;
									if CommandEnabled (T^.Command) then
										if T = Selected then
											Color := CSelect
										else
											Color := CNormal
									else
										if T = Selected then
											Color := CSelDisabled
										else
											Color := CNormDisabled;
									MoveMarkCStr (B [I], T^.Text^, Color, L + 2, SCOff);
								end;
							Inc (I, L + 2);
						end;
				end;
			T := T^.Next;
		end;
	if I < Size.X - 4 then
		begin
			if Seg (Hint) <> 0 then
				HintBuf := Hint (HelpCtx)
			else
				HintBuf := '';
			if HintBuf <> '' then
				begin
					MoveChar (B [I], '', Lo (CNormal), 1);
					Inc (I);
					if I + Length (HintBuf) > Size.X - 4 then HintBuf[0] := Char (Size.X - I - 4);
					MoveCStr (B [I], HintBuf, CNormal);
				end;
		end;
	WriteLine (0, 0, Size.X, 1, B);
end;

procedure TStatusLine.FindItems;
var
	P	: PStatusDef;
begin
	P := Defs;
	while (PtrRec (P).Seg <> 0) and ((HelpCtx < P^.Min) or (HelpCtx > P^.Max)) do
		P := P^.Next;
	if PtrRec (P).Seg = 0 then
		Items := NIL
	else
		Items := P^.Items;
end;

procedure TStatusLine.HandleEvent (var Event : TEvent);
var
	Mouse	: TPoint;
	T		: PStatusItem;

	function ItemMouseIsIn : PStatusItem;
	var
		I,
		K	: Word;
		T	: PStatusItem;
	begin
		ItemMouseIsIn := NIL;
		if Mouse.Y <> 0 then Exit;
		I := 0;
		T := Items;
		while PtrRec (T).Seg <> 0 do
			begin
				if PtrRec (T^.Text).Seg <> 0 then
					begin
						K := I + CStrLen (T^.Text^);
						if K > I then Inc (K, 2);
						if (Mouse.X >= I) and (Mouse.X < K) then
							begin
								ItemMouseIsIn := T;
								Exit;
							end;
						I := K;
					end;
				T := T^.Next;
			end;
	end;

begin
	TView.HandleEvent (Event);
	case Event.What of
		evMouseDown	:
			begin
				T := NIL;
				repeat
					MakeLocal (Event.Where, Mouse);
					if T <> ItemMouseIsIn then
						begin
							T := ItemMouseIsIn;
							DrawSelect (T);
						end;
				until not MouseEvent (Event, evMouseMove);
				if (PtrRec (T).Seg <> 0) and CommandEnabled (T^.Command) then
					SendCommand  (T^.Command);
				DrawView;
			end;
		evKeyDown	:
			begin
				T := Items;
				while PtrRec (T).Seg <> 0 do
					if (Event.KeyCode = T^.KeyCode) and	CommandEnabled (T^.Command) then
						begin
							ClearEvent (Event);
							SendCommand (T^.Command);
							Exit;
						end
					else
						T := T^.Next;
			end;
		evBroadcast	:
			if Event.Command = cmCommandSetChanged then DrawView;
	end;
end;

procedure TStatusLine.Store (var S : TStream);

	procedure DoStoreStatusItems (Cur : PStatusItem);
	var
		T		: PStatusItem;
		Count	: Integer;
	begin
		Count := 0;
		T := Cur;
		while PtrRec (T).Seg <> 0 do
			begin
				Inc (Count);
				T := T^.Next
			end;
		S.Write (Count, SizeOf (Integer));
		while PtrRec (Cur).Seg <> 0 do
			begin
				S.WriteStr (Cur^.Text);
				S.Write (Cur^.KeyCode, SizeOf (Word) * 2);
				Cur := Cur^.Next;
			end;
	end;

	procedure DoStoreStatusDefs (Cur : PStatusDef);
	var
		Count	: Integer;
		T		: PStatusDef;
	begin
		Count := 0;
		T := Cur;
		while PtrRec (T).Seg <> 0 do
			begin
				Inc (Count);
				T := T^.Next
			end;
		S.Write (Count, SizeOf (Integer));
		while PtrRec (Cur).Seg <> 0 do
			begin
				with Cur^ do
					begin
						S.Write (Min, SizeOf (Word) * 2);
						DoStoreStatusItems (Items);
					end;
				Cur := Cur^.Next;
			end;
	end;

begin
	TView.Store (S);
	DoStoreStatusDefs (Defs);
end;

procedure TStatusLine.Update;
var
	H	: Word;
	P	: PView;
begin
	P := TopView;
	if PtrRec (P).Seg <> 0 then
		H := P^.GetHelpCtx
	else
		H := hcNoContext;
	if HelpCtx <> H then
		begin
			HelpCtx := H;
			FindItems;
			DrawView;
		end;
end;

function NewStatusDef (AMin, AMax : Word; AItems : PStatusItem; ANext : PStatusDef) : PStatusDef;
var
	T	: PStatusDef;
begin
	New (T);
	with T^ do
		begin
			Next := ANext;
			Min := AMin;
			Max := AMax;
			Items := AItems;
		end;
	NewStatusDef := T;
end;

function NewStatusKey (AText : String; AKeyCode : Word; ACommand : Word; ANext : PStatusItem) : PStatusItem;
var
	T	: PStatusItem;
begin
	New (T);
	With T^ do
		begin
			Text := GenStr (@AText);
			KeyCode := AKeyCode;
			Command := ACommand;
			Next := ANext;
		end;
	NewStatusKey := T;
end;

procedure RegisterMenus;
begin
	RegisterType (RMenuBar);
	RegisterType (RMenuBox);
	RegisterType (RStatusLine);
end;

end.	{ !!!}
