{Copyright by Leonid Youriev, 1993, os@imfiko.bishkek.su}
{$I VERSION}

unit Number;

interface

uses	Objects, Views, Dialogs, Drivers;

{$I VERSION}

type

	PInputNumber = ^TInputNumber;
	TInputNumber = Object (TInputLine)
		procedure HandleEvent (Var Event : TEvent); virtual;
		procedure SetData (Var Rec); virtual;
		procedure GetData (Var Rec); virtual;
	end;

	PInputInteger = ^TInputInteger;
	TInputInteger = Object (TInputNumber)
		constructor Init (Var R : TRect);
		procedure	GetData (Var Rec); virtual;
		procedure	SetData (Var Rec); virtual;
		function	DataSize : Word; virtual;
		procedure	HandleEvent (Var Event : TEvent); virtual;
	end;

	PInputReal = ^TInputReal;
	TInputReal = Object (TInputNumber)
		constructor Init (Var R : TRect);
		procedure	GetData (Var Rec); virtual;
		procedure	SetData (Var Rec); virtual;
		function	DataSize : Word; virtual;
	end;

	PInputUnsigned = ^TInputUnsigned;
	TInputUnsigned = Object (TInputInteger)
		constructor Init (Var R : TRect);
		procedure	GetData (Var Rec); virtual;
		procedure	SetData (Var Rec); virtual;
		procedure	HandleEvent (Var Event : TEvent); virtual;
	end;

	PInputByte = ^TInputByte;
	TInputByte = Object (TInputUnsigned)
		constructor Init (Var R : TRect);
		procedure	GetData (Var Rec); virtual;
		procedure	SetData (Var Rec); virtual;
		function	DataSize : Word; virtual;
	end;

implementation

uses	CharLog;

{************************************************************ TInputNumber **}
procedure TInputNumber.SetData (Var Rec);
Var
	S	: String absolute Rec;
	I	: Integer;
begin
	AlfaUpStr (S);
	ReplaceChar (S, '.', ',');
	Replace_0 (S);
	TInputLine.SetData (S);
end;

procedure TInputNumber.GetData (Var Rec);
Var
	S	: String absolute Rec;
begin
	TInputLine.GetData (S);
	AlfaUpStr (S);
	ReplaceChar (S, 'O', '0');
	ReplaceChar (S, ',', '.');
	RemoveLowSpace (S);
end;

procedure TInputNumber.HandleEvent (Var Event : TEvent);
Var
	C	: Char;
begin
	if (Event.What = evKeyDown) and (Event.CharCode > ' ') then
		begin
			C := AlfaUpCase (Event.CharCode);
			if C = 'O' then C := '0';
			if C = '.' then C := ',';
			if (C <> ' ') and (C <> ',') and (C <> '-') and ( (C < '0') or (C > '9')) then
				Exit
			else
				Event.CharCode := C;
		end;
	TInputLine.HandleEvent (Event);
end;

{************************************************************** TInputByte **}
constructor TInputByte.Init (Var R : TRect);
begin
	TInputNumber.Init (R, 3);
end;

function TInputByte.DataSize : Word;
begin
	DataSize := SizeOf (Byte);
end;

procedure TInputByte.SetData (Var Rec);
Var
	B	: Byte absolute Rec;
	W	: Word;
begin
	W := B;
	TInputUnsigned.SetData (W);
end;

procedure TInputByte.GetData (Var Rec);
Var
	B	: Byte absolute Rec;
	W	: Word;
begin
	TInputUnsigned.GetData (W);
	B := Lo (W);
end;

{*********************************************************** TInputUnsiged **}
constructor TInputUnsigned.Init (Var R : TRect);
begin
	TInputNumber.Init (R, 5);
end;

procedure TInputUnsigned.SetData (Var Rec);
Var
	W	: Word absolute Rec;
	S	: String [12];
begin
	Str (W, S);
	TInputNumber.SetData (S);
end;

procedure TInputUnsigned.GetData (Var Rec);
Var
	W	: Word absolute Rec;
	S	: String [12];
	N	: Integer;
begin
	TInputNumber.GetData (S);
	Val (S, W, N);
	if N <> 0 then W := 0;
end;

procedure TInputUnsigned.HandleEvent (Var Event : TEvent);
begin
	if (Event.What <> evKeyDown) or (Event.CharCode <> '-') then
		TInputNumber.HandleEvent (Event);
end;

{************************************************************** TInputReal **}
constructor TInputReal.Init (Var R : TRect);
begin
	TInputNumber.Init (R, 12);
end;

function TInputReal.DataSize : Word;
begin
	DataSize := SizeOf (Real);
end;

procedure TInputReal.SetData (Var Rec);
Var
	R	: Real absolute Rec;
	S	: String [24];
begin
	Str (R, S);
	TInputNumber.SetData (S);
end;

procedure TInputReal.GetData (Var Rec);
Var
	R	: Real absolute Rec;
	S	: String [24];
	N	: Integer;
begin
	TInputNumber.GetData (S);
	Val (S, R, N);
	if N <> 0 then R := 0.0;
end;

{*********************************************************** TInputInteger **}
constructor TInputInteger.Init (Var R : TRect);
begin
	TInputNumber.Init (R, 6);
end;

function TInputInteger.DataSize : Word;
begin
	DataSize := SizeOf (Integer);
end;

procedure TInputInteger.SetData (Var Rec);
Var
	I	: Integer absolute Rec;
	S	: String [12];
begin
	Str (I, S);
	TInputNumber.SetData (S);
end;

procedure TInputInteger.GetData (Var Rec);
Var
	I	: Integer absolute Rec;
	S	: String [12];
	N	: Integer;
begin
	TInputNumber.GetData (S);
	Val (S, I, N);
	if N <> 0 then I := 0;
end;

procedure TInputInteger.HandleEvent (Var Event : TEvent);
begin
	if (Event.What <> evKeyDown) or (Event.CharCode <> '.') then
		TInputNumber.HandleEvent (Event);
end;

end.