{Copyright by Leonid Youriev, 1993, os@imfiko.bishkek.su}
{$I VERSION}

unit TabView;

interface

uses	Objects, Drivers, Views, Dialogs, SmartDra;

{$I VERSION}

const

	cfEditable		= $01;
	cfGroup			= $02;
	cfPredSelect	= $04;
	cfNextSelect	= $08;

	cfMobile		= $10;
	cfNumber		= $20;


type

	PColumn = ^TColumn;
	TColumn = Object (TObject)
		Flags		: Byte;
		Size		: Integer;
		Width		: Integer;
		Name		: PString;
		constructor Init (AName : PString; ASize, AWidth, AHeight : Integer; ASub, ANext : PColumn);
		procedure	Title (Var Buffer : TSmartBuffer; Level, Cursor : Integer); virtual;
		procedure	GetView (Var S : String; Var Data); virtual;
		procedure	NextState (By : Boolean; Var Data); virtual;
		function	SetView (S : String; Var Data) : Integer; virtual;
		destructor	Done; virtual;
	private
		Height		: Word;
		Next		: PColumn;
		Sub			: PColumn;
		NextView	: PColumn;
		function	AdjustOfs (AOfs : Word) : Word;
		function	GetLevels : Integer;
	end;

	PTitle = ^TTitle;
	TTitle = Object (TColumn)
		Levels	: Integer;
		constructor Init (ASub : PColumn);
		function	Find (Var AView : Integer) : PColumn;
	end;

	PLeftTitle = ^TLeftTitle;
	TLeftTitle = Object (TTitle)
		procedure	Title (Var Buffer : TSmartBuffer; Level, Cursor : Integer); virtual;
	end;

	PTabViewer = ^TTabViewer;
	TTabViewer = Object (TListViewer)
		Shift		: Integer;
		Field		: PColumn;
		Edit		: String;
		Undo		: String;
		Title		: PTitle;
		constructor	Init (Var R : TRect; HS, VS : PScrollBar; ATitle : PTitle);
		constructor	Load (Var S : TStream);
		procedure	Store (Var S : TStream);
		function	Data (I : Integer) : Pointer; virtual;
		procedure	Modified (I : Integer); virtual;
		procedure	NewField (F : PColumn); virtual;
		procedure	NewLine (I : Integer); virtual;
		procedure	HandleEvent (Var Event : TEvent); virtual;
		procedure	FocusItem (Item : Integer); virtual;
		function	Seek (X, Y : Integer) : Boolean; virtual;
		function	PosX : Integer;
		function	Flush : Boolean;
		procedure	ChangeBounds (Var B : TRect); virtual;
		procedure	ResetRange (I : Integer);
		procedure	UpDateCommands; virtual;
		procedure	PrintStr (Var Buffer : TSmartBuffer; Line : Integer);
	private
		Modif		: Boolean;
		procedure	DrawLine (Var B : TDrawBuffer; P : PColumn; D : Pointer; CL, CI, CR : Byte; W, N : Integer);
		procedure	Draw; virtual;
	end;

	PHeader = ^THeader;
	THeader = Object (TStaticText)
	end;

const
	RTabViewer	: TStreamRec = (
		ObjType	: 2003;
		VmtLink	: Ofs(TypeOf(TTabViewer)^);
		Load	: @TTabViewer.Load;
		Store	: @TTabViewer.Store );

Var
	Clip	: String;

implementation

uses	CharLog, SysVars;

procedure NStr (Var S : String; I : Integer); assembler;
asm
	LES		DI, S
	CLD
	MOV		CL, ES:[DI]
	MOV		AL, BYTE PTR I
	STOSB
	CMP		CL, AL
	JAE		@1
	MOV		CH, 0
	ADD		DI, CX
	SUB		CL, AL
	NEG		CL
	MOV		AL, ' '
	REP		STOSB
@1:
end;

{************************************************************** TLeftTitle **}
procedure TLeftTitle.Title (Var Buffer : TSmartBuffer; Level, Cursor : Integer);
begin
	Buffer.AddChar ('');
	TTitle.Title (Buffer, Level, Cursor);
end;

{****************************************************************** TTitle **}
constructor TTitle.Init (ASub : PColumn);
Var
	I,
	J	: Integer;
	P,
	N	: PColumn;
begin
	TColumn.Init (NIL, 0, 0, 1, ASub, NIL);
	Size := AdjustOfs (0);
	Levels := GetLevels + 1;
	I := 0;
	P := Find (I);
	while PtrRec (P).Seg > 0  do
		begin
			Inc (I, P^.Width);
			J := I;
			N := Find (J);
			P^.NextView := N;
			P := N;
		end;
end;

function TTitle.Find (Var AView : Integer) : PColumn; assembler;
asm
	MOV		BX, DS
	LDS		DI, AView
	MOV		SI, [DI]
	LDS		DI, Self
@1:	MOV		AX, [DI].Width
	CMP		SI, AX
	JL		@3
	SUB		SI, AX
	LDS		DI, [DI].Next
	MOV		AX, DS
	OR		AX, AX
	JNE		@1
	JMP		@4
@3:	CMP		WORD PTR [DI].Sub + 2, 0
	JE		@4
	LDS		DI, [DI].Sub
	JMP		@1
@4:	MOV		DX, DS
	MOV		AX, DI
	LDS		DI, AView
	MOV		[DI], SI
	MOV		DS, BX
end;

{***************************************************************** TColumn **}
constructor TColumn.Init (AName : PString; ASize, AWidth, AHeight : Integer; ASub, ANext : PColumn);
begin
	TObject.Init;
	Name := AName;
	Next := ANext;
	Sub := ASub;
	if PtrRec (ASub).Seg > 0 then
		begin
			Flags := cfGroup;
			ASize := 0;
			AWidth := 0;
			repeat
				Inc (ASize, ASub^.Size);
				Inc (AWidth, ASub^.Width);
				ASub := ASub^.Next;
			until PtrRec (ASub).Seg = 0;
		end
	else
		begin
			Flags := cfEditable;
			Inc (AWidth);
		end;
	Size := ASize;
	Width := AWidth;
	Height := AHeight + 1;
end;

procedure TColumn.GetView (Var S : String; Var Data);
begin
end;

function TColumn.SetView (S : String; Var Data) : Integer;
begin
	SetView := 0;
end;

procedure TColumn.NextState (By : Boolean; Var Data);
begin
end;

procedure TColumn.Title (Var Buffer : TSmartBuffer; Level, Cursor : Integer);
Var
	I,
	M,
	W,
	B	: Integer;
begin
	W := Width - 1;
	if Level < 0 then
		if PtrRec (Sub).Seg > 0 then
			Sub^.Title (Buffer, -1, -1)
		else
			begin
				Buffer.RepeatChar (W, '');
				Buffer.AddChar ('');
			end
	else
		if Level = 0 then
			begin
				Buffer.RepeatChar (W, '');
				Buffer.AddChar ('');
			end
	else
		if (Level < Height) or (PtrRec (Sub).Seg = 0) then
			begin
				I := Length (Name^);
				B := W * (Level - 1);
				if B < I then
					if W + B > I then
						M := I - B
					else
						M := W
				else
					M := 0;
				I := (W - M) div 2;
				Buffer.Pure (I);
				if (Cursor >= 0) and (Cursor < W) then Buffer.AddChar ('~');
				Buffer.AddClip (Name^, B, M);
				if (Cursor >= 0) and (Cursor < W) then Buffer.AddChar ('~');
				Buffer.Pure (W - M - I);
				Buffer.AddChar ('');
			end
		else
			Sub^.Title (Buffer, Level - Height, Cursor);
	if (PtrRec (Next).Seg > 0) and Buffer.Empty then
		Next^.Title (Buffer, Level, Cursor - Width);
end;

function TColumn.AdjustOfs (AOfs : Word) : Word;
begin
	if PtrRec (Sub).Seg > 0 then
		Size := Sub^.AdjustOfs (AOfs)
	else
		PtrRec (Sub).Ofs := AOfs;
	if PtrRec (Next).Seg > 0 then
		AdjustOfs := Size + Next^.AdjustOfs (AOfs + Size)
	else
		AdjustOfs := Size;
end;

function TColumn.GetLevels : Integer;
Var
	LN		: Integer;
	Level	: Integer;
	P		: PColumn;
begin
	if PtrRec (Sub).Seg > 0 then
		Level := Height + Sub^.GetLevels - 1
	else
		Level := Height - 1;
	if (PtrRec (Next).Seg > 0) then
		LN := Next^.GetLevels
	else
		LN := 0;
	if Level > LN then
		GetLevels := Level
	else
		GetLevels := LN;
end;

destructor TColumn.Done;
begin
	Next^.Free;
	Sub^.Free;
	DisposeStr (Name);
end;

{************************************************************** TTabViewer **}
constructor TTabViewer.Init (Var R : TRect; HS, VS : PScrollBar; ATitle : PTitle);
Var
	M	: Integer;
begin
	TListViewer.Init (R, 1, HS, VS);
	Title := ATitle;
	if PtrRec (HS).Seg > 0 then
		HS^.SetRange (0, ATitle^.Width);
	if PtrRec (VS).Seg > 0 then
		VS^.SetParams (0, 0, 0, Range, Size.Y - Title^.Levels, 1);
	ResetRange (0);
end;

procedure TTabViewer.ChangeBounds (Var B : TRect);
Var
	E	: Word;
begin
	E := EventMask;
	EventMask := 0;
	TView. ChangeBounds (B);
	FocusItem (Focused);
	if PtrRec (VScrollBar).Seg > 0 then
		begin
			VScrollBar^.SetStep (Size.Y - Title^.Levels, 1);
			VScrollBar^.DrawView;
		end;
	if PtrRec (HScrollBar).Seg > 0 then
		begin
			HScrollBar^.SetStep (Size.X, 1);
			HScrollBar^.DrawView;
		end;
	EventMask := E;
end;

procedure TTabViewer.HandleEvent (Var Event : TEvent);
Var
	X,
	V	: Integer;
	P	: PColumn;
	B	: Boolean;
Label
	No;
	function InLog : Boolean;
		begin
			B := True;
			InLog := False;
			if (Options and ofEditable = 0) or (Focused >= Range) then Exit;
			X := PosX; V := X;
			P := Title^.Find (V); Inc (V);
			if (PtrRec (P).Seg = 0) or (P^.Flags and cfEditable = 0) then Exit;
			InLog := True;
		end;

	procedure OutLog;
		begin
			if (P^.Flags and cfMobile <> 0) and B then
			begin
				P^.SetView (Edit, Pointer (Longint (Data (Focused)) + PtrRec (P^.Sub).Ofs)^);
				P^.GetView (Edit, Pointer (Longint (Data (Focused)) + PtrRec (P^.Sub).Ofs)^);
			end;
			NStr (Edit, P^.Width - 1);
			Modif := Undo <> Edit;
			if not Seek (X, Focused) then DrawView;
		end;
begin
	case Event.What of
		evKeyDown	:
			begin
				case Event.KeyCode of
					kbLeft		: Seek (PosX - 1, Focused);
					kbRight		: Seek (PosX + 1, Focused);
					kbIns		: SetState (sfCursorIns, State and sfCursorIns = 0);
					kbHome		: Seek (0, Focused);
					kbEnd		: Seek (MaxInt, Focused);
					kbCtrlIns	: Clip := Edit;
				else
					if not InLog then goto No;
				if P^.Flags and cfMobile = 0 then
					case Event.KeyCode of
						kbBack,
						kbDel		:
							begin
								if Event.KeyCode = kbBack then
									begin
										Dec (X);
										Dec (V);
									end;
								if V > 0 then System.Delete (Edit, V, 1);
							end;
						kbCtrlBack	: CutSpace (Edit);
						kbCtrlDel	: SendCommand (cmDel);
						kbShiftDel	: SendCommand (cmCut);
						kbShiftIns	: SendCommand (cmPaste);
						kbCtrlIns	: SendCommand (cmCopy);
					else
						if Event.CharCode = ^U then
							SendCommand (cmUndo)
						else
							if Event.CharCode >= ' ' then
								begin
									if Length (Edit) < V then NStr (Edit, V);
									if State and sfCursorIns = 0 then
										System.Insert (Event.CharCode, Edit, V)
									else
										Edit [V] := Event.CharCode;
									Inc (X);
								end
							else
								goto No;
					end
				else
					if Event.CharCode = ' ' then
						begin
							P^.NextState (not ShiftPressed, Pointer (Longint (Data (Focused)) + PtrRec (P^.Sub).Ofs)^);
							P^.GetView (Edit, Pointer (Longint (Data (Focused)) + PtrRec (P^.Sub).Ofs)^);
							B := False;
						end
					else
						if Event.CharCode = ^U then
							SendCommand (cmUndo)
						else
							goto No;
				end;
				OutLog;
			end;
		evCommand	:
			begin
				if not InLog then Exit;
				case Event.Command of
					cmDel:	Edit := '';
					cmCut:	begin
								Clip := Edit;
								Edit := '';
							end;
					cmCopy: Clip := Edit;
					cmPaste:if Clip <> '' then Edit := Clip;
					cmUndo:	Edit := Undo;
				else
					goto No;
				end;
				OutLog;
			end;
	else
		if (Event.What = evBroadCast) and (Event.InfoPtr = @Self) then
			begin
				if Event.Command = cmReceivedFocus then UpDateCommands;
				Exit;
			end
		else
			if Event.Command = cmDefault then
				begin
					if Flush then
						begin
							if (PtrRec (Field).Seg > 0) and (Focused < Range) then
								Field^.GetView (Edit, Pointer (Longint (Data (Focused)) + PtrRec (Field^.Sub).Ofs)^)
							else
								Edit := '';
							Undo := Edit;
						end;
					DrawView;
				end
			else
				begin
					No:
					TListViewer.HandleEvent (Event);
					Exit;
				end;
	end;
	ClearEvent (Event);
end;

procedure TTabViewer.UpDateCommands;
begin
	DisableCommands ([cmCopy, cmDel, cmCut, cmUndo, cmPaste, cmAgain, cmReplace, cmClear, cmAgain, cmNew, cmEdit, cmFind]);
	if Range > 0 then
		if Options and ofEditable <> 0 then
			EnableCommands ([cmCopy, cmDel, cmCut, cmUndo, cmPaste, cmClear, cmEdit, cmNew, cmReplace, cmFind, cmAgain])
		else
			EnableCommands ([cmCopy, cmFind, cmAgain]);
end;

procedure TTabViewer.PrintStr (Var Buffer : TSmartBuffer; Line : Integer);
Var
	P	: PColumn;
	D	: Pointer;
	I	: Integer;
	S	: String;
begin
	if (Line < 0) or (Line > Range + Title^.Levels) then Exit;
	if Line < Title^.Levels then
		Title^.Title (Buffer, Line + 2, -1)
	else
		if Line = Range + Title^.Levels then
			Title^.Title (Buffer, -1, -1)
		else
			begin
				I := 0;
				P := Title^.Find (I);
				D := Data (Line - Title^.Levels);
				if PtrRec (D).Seg = 0 then Exit;
				while PtrRec (P).Seg > 0 do
					begin
						P^.GetView (S, Ptr (PtrRec (D).Seg, PtrRec (D).Ofs + PtrRec (P^.Sub).Ofs)^);
						NStr (S, P^.Width - 1);
						Buffer.AddClip (S, 0, P^.Width - 1);
						Buffer.AddChar ('');
						P := P^.NextView;
					end;
			end;
end;

procedure TTabViewer.DrawLine (Var B : TDrawBuffer; P : PColumn; D : Pointer; CL, CI, CR : Byte; W, N : Integer);
Var
	O	: Integer;
	S	: String;
	C	: Char;
begin
	O := 0;
	if PtrRec (D).Seg > 0 then
		while (W > O) and (PtrRec (P).Seg > 0) do
			begin
				C := '';
				if (N = Focused) and (P = Field) then
					begin
						MoveStr (B [O], Edit, CI);
						if Modif then C := '';
					end
				else
					begin
						P^.GetView (S, Ptr (PtrRec (D).Seg, PtrRec (D).Ofs + PtrRec (P^.Sub).Ofs)^);
						NStr (S, P^.Width - 1);
						if (N = Focused) or (P = Field) then
							MoveStr (B [O], S, CL)
						else
							MoveStr (B [O], S, CR);
					end;
				Inc (O, P^.Width);
				MoveChar (B [O - 1], C, CR, 1);
				P := P^.NextView;
			end;
	if (O < W) and (PtrRec (Owner).Seg > 0) then MoveChar (B [O], ' ', GetColor (6), W - O);
end;

procedure TTabViewer.Draw;
Var
	Sf1		: TDrawBuffer;
	B		: TDrawBuffer;
	Y,
	N,
	V,
	W		: Integer;
	NC,
	SC,
	FC,
	MC,
	IC,
	OC		: Byte;
	CR		: Word;
	CL,
	CI 		: Byte;
	P		: PColumn;
	Buffer	: TSmartDraw;
begin
	N := TopItem;
	NC := Lo (GetColor ($0F));
	SC := Lo (GetColor ($10));
	MC := Lo (GetColor ($25));
	if PtrRec (Owner).Seg > 0 then OC := Lo (Owner^.GetColor (1));
	WordRec (CR).Lo := NC;
	if State and (sfSelected + sfActive) = (sfSelected + sfActive) then
		begin
			IC := Lo(GetColor (11));
			FC := Lo (GetColor (12));
			WordRec (CR).Hi := FC;
		end
	else
		begin
			IC := Lo (GetColor (13));
			FC := Lo (GetColor (13));
			WordRec (CR).Hi := NC;
		end;
	Buffer.Init (Size.X + 1, CR, OC);
	Buffer.Clear (Shift);
	V := Shift;
	P := Title^.Find (V);
	W := Size.X + V;
	for Y := 0 to Size.Y do
		if Y < Title^.Levels then
			begin
				Title^.Title (Buffer, Y + 2, PosX);
				Buffer.Adjust;
				WriteLine (0, Y, Size.X, 1, Buffer.Draw^);
				Buffer.Swap;
			end
		else
			if N < Range then
				begin
					if N = Focused then
						begin
							CI := IC;
							CL := FC;
						end
					else
						if N = Marker then
							begin
								CI := FC;
								CL := MC;
							end
						else
							if IsSelected (N) then
								begin
									CI := NC;
									CL := SC;
								end
							else
								begin
									CI := FC;
									CL := FC;
								end;
					DrawLine (B, P, Data (N), CL, CI, NC, W, N);
					WriteLine (0, Y, Size.X, 1, B [V]);
					Inc (N);
				end
			else
				if N = Range then
					begin
						Title^.Title (Buffer, -1, -1);
						Buffer.Adjust;
						WriteLine (0, Y, Size.X, 1, Buffer.Draw^);
						Inc (N);
					end
				else
					begin
						MoveChar (B, ' ', OC, Size.X);
						WriteLine (0, Y, Size.X, 1, B);
					end;
end;

procedure TTabViewer.Modified (I : Integer);
begin
end;

procedure TTabViewer.NewField (F : PColumn);
begin
end;

procedure TTabViewer.NewLine (I : Integer);
begin
end;

function TTabViewer.Data (I : Integer) : Pointer;
begin
	Data := NIL;
end;

function TTabViewer.PosX : Integer;
begin
	PosX := Shift + Cursor.X;
end;

procedure TTabViewer.ResetRange (I : Integer);
Var
	Y	: Integer;
	E	: Word;
begin
	E := EventMask;
	EventMask := evNothing;
	if (PtrRec (VScrollBar).Seg > 0) and (VScrollBar^.Max <> I) then
		VScrollBar^.SetRange (0, I);
	SetRange (I);
	Modif := False;
	Y := Focused; Focused := -2;
	if not Seek (PosX, Y) then DrawView;
	UpDateCommands;
	EventMask := E;
end;

procedure TTabViewer.FocusItem (Item : Integer);
begin
	Seek (PosX, Item);
end;

function TTabViewer.Flush : Boolean;
Var
	V,
	I	: Integer;
begin
	if Modif and (PtrRec (Field).Seg <> 0) then
		V := Field^.SetView (Edit, Pointer (Longint (Data (Focused)) + PtrRec (Field^.Sub).Ofs)^)
	else
		V := 0;
	if V = 0 then
		begin
			if Modif then Modified (Focused);
			Modif := False;
			Flush := True;
		end
	else
		begin
			I := PosX;
			Title^.Find (I);
			Seek (PosX - I + V - 1, Focused);
			Flush := False;
			Exit;
		end;
end;

function TTabViewer.Seek (X, Y : Integer) : Boolean;
Var
	F		: PColumn;
	V,
	S,
	I		: Integer;
	ReDraw	: Boolean;
	T		: Integer;
	N		: Boolean;
	E		: Word;
	Q		: Integer;
Label
	St;
begin
	T := 0;
	Q := 0;
	if Y >= Range then
		begin
			Y := Range - 1;
			if TopItem - Title^.Levels + Size.Y >= Range then Q := 1;
		end;
	if Y < 0 then Y := 0;
	S := Shift;
	St:
	if X >= 0 then
		begin
			V := X;
			F := Title^.Find (V);
			if (PtrRec (F).Seg > 0) and (V = F^.Width - 1) then
				if PosX > X then
					Dec (X)
				else
					begin
						Inc (X);
						goto St;
					end;
		end
	else
		F := Field;
	N := (F <> Field) or (Y <> Focused);
	if N then
		if not Flush then
			begin
				Seek := False;
				Exit;
			end
		else
			ReDraw := True
	else
		ReDraw := False;
	if X < 0 then
		begin
			X := 0;
			if Title^.Flags and cfPredSelect <> 0 then T := -1;
		end;
	if X >= Title^.Width then
		begin
			if (Shift + Size.X >= Title^.Width) and (Title^.Flags and cfNextSelect <> 0) then T := 1;
			X := Title^.Width - 1;
		end;
	if S > X then S := X;
	if X - S >= Size.X then
		S := X - Size.X + 1;
	E := EventMask;
	EventMask := evNothing;
	if S <> Shift then
		begin
			ReDraw := True;
			Shift := S;
			if PtrRec (HScrollBar).Seg > 0 then
				HScrollBar^.SetValue (S);
		end;
	if X = Title^.Width - 1 then Dec (X);
	if PtrRec (F).Seg = 0 then
		begin
			V := X;
			F := Title^.Find (V);
		end;
	if F <> Field then
		begin
			NewField (F);
			Field := F;
		end;
	if N then
		begin
			if (Y < Range) and (PtrRec (F).Seg > 0) then
				F^.GetView (Edit, Pointer (Longint (Data (Y)) + PtrRec (F^.Sub).Ofs)^)
			else
				Edit := '';
			Undo := Edit;
		end;
	if (Y + Q <> Focused)  then
		begin
			ReDraw := True;
			NewLine (Y);
			V := TopItem;
			Focused := Y;
			if V > Y then V := Y;
			Inc (Y, Q);
			if Y - V + Title^.Levels >= Size.Y then
				V := Y - Size.Y + Title^.Levels + 1;
			Dec (Y, Q);
			if TopItem <> V then
				begin
					TopItem := V;
					if PtrRec (VScrollBar).Seg > 0 then
						VScrollBar^.SetValue (V);
				end;
		end;
	if N then HideCursor;
	SetCursor (X - S, Y - TopItem + Title^.Levels);
	EventMask := E;
	if ReDraw and (T = 0) then DrawView;
	if (T <> 0) and (PtrRec (Owner).Seg > 0) then Owner^.SelectNext (T < 0);
	Seek := ReDraw or (T <> 0);
	SetState (sfCursorVis, Y < Range);
end;

constructor TTabViewer.Load (Var S : TStream);
begin
	TListViewer.Load (S);
	if Options and ofEditable <> 0 then
		S.Read (Shift, SizeOf (Shift) + SizeOf (Field) + SizeOf (Edit) + SizeOf (Undo));
end;

procedure TTabViewer.Store (Var S : TStream);
begin
	TListViewer.Store (S);
	if Options and ofEditable <> 0 then
		S.Write (Shift, SizeOf (Shift) + SizeOf (Field)	+ SizeOf (Edit) + SizeOf (Undo));
end;

end.